# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import abstractmethod

from gravitino.api.rel.expressions.expression import Expression


class FunctionExpression(Expression):
    """
    The interface of a function expression. A function expression is an expression that takes a
    function name and a list of arguments.
    """

    @staticmethod
    def of(function_name: str, *arguments: Expression) -> FuncExpressionImpl:
        """
        Creates a new FunctionExpression with the given function name.
        If no arguments are provided, it uses an empty expression.

        :param function_name: The name of the function.
        :param arguments: The arguments to the function (optional).
        :return: The created FunctionExpression.
        """
        arguments = list(arguments) if arguments else Expression.EMPTY_EXPRESSION
        return FuncExpressionImpl(function_name, arguments)

    @abstractmethod
    def function_name(self) -> str:
        """Returns the function name."""

    @abstractmethod
    def arguments(self) -> list[Expression]:
        """Returns the arguments passed to the function."""

    def children(self) -> list[Expression]:
        """Returns the arguments as children."""
        return self.arguments()


class FuncExpressionImpl(FunctionExpression):
    """
    A concrete implementation of the FunctionExpression interface.
    """

    _function_name: str
    _arguments: list[Expression]

    def __init__(self, function_name: str, arguments: list[Expression]):
        super().__init__()
        self._function_name = function_name
        self._arguments = arguments

    def function_name(self) -> str:
        return self._function_name

    def arguments(self) -> list[Expression]:
        return self._arguments

    def __str__(self) -> str:
        if not self._arguments:
            return f"{self._function_name}()"
        arguments_str = ", ".join(map(str, self._arguments))
        return f"{self._function_name}({arguments_str})"

    def __eq__(self, other: FuncExpressionImpl) -> bool:
        if self is other:
            return True
        if other is None or self.__class__ is not other.__class__:
            return False
        return (
            self._function_name == other.function_name()
            and self._arguments == other.arguments()
        )

    def __hash__(self) -> int:
        return hash((self._function_name, tuple(self._arguments)))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from typing import Dict, List

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.api.job.job_handle import JobHandle
from gravitino.api.job.job_template import JobTemplate
from gravitino.api.job.job_template_change import JobTemplateChange
from gravitino.api.job.supports_jobs import SupportsJobs
from gravitino.api.tag.tag import Tag
from gravitino.api.tag.tag_operations import TagOperations
from gravitino.client.dto_converters import DTOConverters
from gravitino.client.generic_job_handle import GenericJobHandle
from gravitino.client.generic_tag import GenericTag
from gravitino.dto.metalake_dto import MetalakeDTO
from gravitino.dto.requests.catalog_create_request import CatalogCreateRequest
from gravitino.dto.requests.catalog_set_request import CatalogSetRequest
from gravitino.dto.requests.catalog_updates_request import CatalogUpdatesRequest
from gravitino.dto.requests.job_run_request import JobRunRequest
from gravitino.dto.requests.job_template_register_request import (
    JobTemplateRegisterRequest,
)
from gravitino.dto.requests.job_template_updates_request import (
    JobTemplateUpdatesRequest,
)
from gravitino.dto.requests.tag_create_request import TagCreateRequest
from gravitino.dto.responses.catalog_list_response import CatalogListResponse
from gravitino.dto.responses.catalog_response import CatalogResponse
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.job_list_response import JobListResponse
from gravitino.dto.responses.job_response import JobResponse
from gravitino.dto.responses.job_template_list_response import JobTemplateListResponse
from gravitino.dto.responses.job_template_response import JobTemplateResponse
from gravitino.dto.responses.tag_response import TagNamesListResponse, TagResponse
from gravitino.exceptions.handlers.catalog_error_handler import CATALOG_ERROR_HANDLER
from gravitino.exceptions.handlers.job_error_handler import JOB_ERROR_HANDLER
from gravitino.exceptions.handlers.tag_error_handler import TAG_ERROR_HANDLER
from gravitino.rest.rest_utils import encode_string
from gravitino.utils.http_client import HTTPClient
from gravitino.utils.precondition import Precondition

logger = logging.getLogger(__name__)


class GravitinoMetalake(
    MetalakeDTO,
    SupportsJobs,
    TagOperations,
):
    """
    Gravitino Metalake is the top-level metadata repository for users. It contains a list of catalogs
    as sub-level metadata collections. With GravitinoMetalake, users can list, create, load,
    alter and drop a catalog with specified identifier.
    """

    rest_client: HTTPClient

    API_METALAKES_CATALOGS_PATH = "api/metalakes/{}/catalogs/{}"
    API_METALAKES_JOB_TEMPLATES_PATH = "api/metalakes/{}/jobs/templates"
    API_METALAKES_JOB_RUNS_PATH = "api/metalakes/{}/jobs/runs"
    API_METALAKES_TAG_PATH = "api/metalakes/{}/tags/{}"
    API_METALAKES_TAGS_PATH = "api/metalakes/{}/tags"

    def __init__(self, metalake: MetalakeDTO = None, client: HTTPClient = None):
        super().__init__(
            _name=metalake.name(),
            _comment=metalake.comment(),
            _properties=metalake.properties(),
            _audit=metalake.audit_info(),
        )
        self.rest_client = client

    def list_catalogs(self) -> List[str]:
        """List all the catalogs under this metalake.

        Raises:
            NoSuchMetalakeException if the metalake with specified namespace does not exist.

        Returns:
            A list of the catalog names under this metalake.
        """
        url = f"api/metalakes/{encode_string(self.name())}/catalogs"
        response = self.rest_client.get(url, error_handler=CATALOG_ERROR_HANDLER)
        entity_list = EntityListResponse.from_json(response.body, infer_missing=True)
        entity_list.validate()
        return [identifier.name() for identifier in entity_list.identifiers()]

    def list_catalogs_info(self) -> List[Catalog]:
        """List all the catalogs with their information under this metalake.

        Raises:
            NoSuchMetalakeException if the metalake with specified namespace does not exist.

        Returns:
            A list of Catalog under the specified namespace.
        """
        params = {"details": "true"}
        url = f"api/metalakes/{encode_string(self.name())}/catalogs"
        response = self.rest_client.get(
            url, params=params, error_handler=CATALOG_ERROR_HANDLER
        )
        catalog_list = CatalogListResponse.from_json(response.body, infer_missing=True)

        return [
            DTOConverters.to_catalog(self.name(), catalog, self.rest_client)
            for catalog in catalog_list.catalogs()
        ]

    def load_catalog(self, name: str) -> Catalog:
        """Load the catalog with specified name.

        Args:
            name: The name of the catalog to load.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.

        Returns:
            The Catalog with specified name.
        """
        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        response = self.rest_client.get(url, error_handler=CATALOG_ERROR_HANDLER)
        catalog_resp = CatalogResponse.from_json(response.body, infer_missing=True)

        return DTOConverters.to_catalog(
            self.name(), catalog_resp.catalog(), self.rest_client
        )

    def create_catalog(
        self,
        name: str,
        catalog_type: Catalog.Type,
        provider: str,
        comment: str,
        properties: Dict[str, str],
    ) -> Catalog:
        """Create a new catalog with specified name, catalog type, comment and properties.

        Args:
            name: The name of the catalog.
            catalog_type: The type of the catalog.
            provider: The provider of the catalog. This parameter can be None if the catalog
            provides a managed implementation. Currently, the model and fileset catalog support
            None provider. For the details, please refer to the Catalog.Type.
            comment: The comment of the catalog.
            properties: The properties of the catalog.

        Raises:
            NoSuchMetalakeException if the metalake does not exist.
            CatalogAlreadyExistsException if the catalog with specified name already exists.

        Returns:
            The created Catalog.
        """

        catalog_create_request = CatalogCreateRequest(
            name=name,
            catalog_type=catalog_type,
            provider=provider,
            comment=comment,
            properties=properties,
        )
        catalog_create_request.validate()

        url = f"api/metalakes/{encode_string(self.name())}/catalogs"
        response = self.rest_client.post(
            url, json=catalog_create_request, error_handler=CATALOG_ERROR_HANDLER
        )
        catalog_resp = CatalogResponse.from_json(response.body, infer_missing=True)

        return DTOConverters.to_catalog(
            self.name(), catalog_resp.catalog(), self.rest_client
        )

    def alter_catalog(self, name: str, *changes: CatalogChange) -> Catalog:
        """Alter the catalog with specified name by applying the changes.

        Args:
            name: the name of the catalog.
            changes: the changes to apply to the catalog.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.
            IllegalArgumentException if the changes are invalid.

        Returns:
            the altered Catalog.
        """

        reqs = [DTOConverters.to_catalog_update_request(change) for change in changes]
        updates_request = CatalogUpdatesRequest(reqs)
        updates_request.validate()

        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        response = self.rest_client.put(
            url, json=updates_request, error_handler=CATALOG_ERROR_HANDLER
        )
        catalog_response = CatalogResponse.from_json(response.body, infer_missing=True)
        catalog_response.validate()

        return DTOConverters.to_catalog(
            self.name(), catalog_response.catalog(), self.rest_client
        )

    def drop_catalog(self, name: str, force: bool = False) -> bool:
        """Drop the catalog with specified name.

        Args:
            name: the name of the catalog.
            force: whether to force drop the catalog.

        Returns:
            true if the catalog is dropped successfully, false if the catalog does not exist.
        """
        params = {"force": str(force)}
        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        response = self.rest_client.delete(
            url, params=params, error_handler=CATALOG_ERROR_HANDLER
        )

        drop_response = DropResponse.from_json(response.body, infer_missing=True)
        drop_response.validate()

        return drop_response.dropped()

    def enable_catalog(self, name: str):
        """Enable the catalog with specified name. If the catalog is already in use, this method does nothing.

        Args:
            name: the name of the catalog.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.
        """

        catalog_enable_request = CatalogSetRequest(in_use=True)
        catalog_enable_request.validate()

        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        self.rest_client.patch(
            url, json=catalog_enable_request, error_handler=CATALOG_ERROR_HANDLER
        )

    def disable_catalog(self, name: str):
        """Disable the catalog with specified name. If the catalog is already disabled, this method does nothing.

        Args:
            name: the name of the catalog.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.
        """

        catalog_disable_request = CatalogSetRequest(in_use=False)
        catalog_disable_request.validate()

        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        self.rest_client.patch(
            url, json=catalog_disable_request, error_handler=CATALOG_ERROR_HANDLER
        )

    ##########
    # Job operations
    ##########

    def list_job_templates(self) -> List[JobTemplate]:
        """List all the registered job templates in Gravitino.

        Returns:
            List of job templates.
        """
        params = {"details": "true"}
        url = self.API_METALAKES_JOB_TEMPLATES_PATH.format(encode_string(self.name()))
        response = self.rest_client.get(
            url, params=params, error_handler=JOB_ERROR_HANDLER
        )
        resp = JobTemplateListResponse.from_json(response.body, infer_missing=True)
        resp.validate()

        return [
            DTOConverters.from_job_template_dto(dto) for dto in resp.job_templates()
        ]

    def register_job_template(self, job_template: JobTemplate) -> None:
        """Register a job template with the specified job template to Gravitino. The registered
        job template will be maintained in Gravitino, allowing it to be executed later.

        Args:
            job_template: The job template to register.

        Raises:
            JobTemplateAlreadyExists: If a job template with the same name already exists.
        """
        url = self.API_METALAKES_JOB_TEMPLATES_PATH.format(encode_string(self.name()))
        req = JobTemplateRegisterRequest(
            DTOConverters.to_job_template_dto(job_template)
        )

        self.rest_client.post(url, json=req, error_handler=JOB_ERROR_HANDLER)

    def get_job_template(self, job_template_name: str) -> JobTemplate:
        """Retrieves a job template by its name.

        Args:
            job_template_name: The name of the job template to retrieve.

        Returns:
            The job template if found, otherwise raises an exception.

        Raises:
            NoSuchJobTemplateException: If no job template with the specified name exists.
        """
        if not job_template_name or not job_template_name.strip():
            raise ValueError("Job template name cannot be null or empty")

        url = (
            f"{self.API_METALAKES_JOB_TEMPLATES_PATH.format(encode_string(self.name()))}/"
            f"{encode_string(job_template_name)}"
        )
        response = self.rest_client.get(url, error_handler=JOB_ERROR_HANDLER)
        resp = JobTemplateResponse.from_json(response.body, infer_missing=True)
        resp.validate()

        return DTOConverters.from_job_template_dto(resp.job_template())

    def delete_job_template(self, job_template_name: str) -> bool:
        """
        Deletes a job template by its name. This will remove the job template from Gravitino, and it
        will no longer be available for execution. Only when all the jobs associated with this job
        template are completed, failed or cancelled, the job template can be deleted successfully,
        otherwise it will throw InUseException. Returns false if the job template to be deleted does
        not exist.

        The deletion of a job template will also delete all the jobs associated with this template.

        Args:
            job_template_name: The name of the job template to delete.

        Returns:
            bool: True if the job template was deleted successfully, False if the job template
            does not exist.

        Raises:
            InUseException: If the job template is currently in use by any jobs, it cannot be deleted.
        """
        if not job_template_name or not job_template_name.strip():
            raise ValueError("Job template name cannot be null or empty")

        url = (
            f"{self.API_METALAKES_JOB_TEMPLATES_PATH.format(encode_string(self.name()))}/"
            f"{encode_string(job_template_name)}"
        )
        response = self.rest_client.delete(url, error_handler=JOB_ERROR_HANDLER)

        drop_response = DropResponse.from_json(response.body, infer_missing=True)
        drop_response.validate()

        return drop_response.dropped()

    def alter_job_template(
        self, job_template_name: str, *changes: JobTemplateChange
    ) -> JobTemplate:
        """Alter the job template with specified name by applying the changes.

        Args:
            job_template_name: the name of the job template.
            changes: the changes to apply to the job template.

        Raises:
            NoSuchJobTemplateException if the job template with specified name does not exist.
            IllegalArgumentException if the changes are invalid.

        Returns:
            the altered JobTemplate.
        """

        reqs = [
            DTOConverters.to_job_template_update_request(change) for change in changes
        ]
        updates_request = JobTemplateUpdatesRequest(reqs)

        url = (
            f"{self.API_METALAKES_JOB_TEMPLATES_PATH.format(encode_string(self.name()))}/"
            f"{encode_string(job_template_name)}"
        )
        response = self.rest_client.put(
            url, json=updates_request, error_handler=JOB_ERROR_HANDLER
        )
        job_template_response = JobTemplateResponse.from_json(
            response.body, infer_missing=True
        )
        job_template_response.validate()

        return DTOConverters.from_job_template_dto(job_template_response.job_template())

    def list_jobs(self, job_template_name: str = None) -> List[JobHandle]:
        """List all the jobs under this metalake.

        Args:
            job_template_name: The name of the job template to filter jobs by. If None, all jobs are listed.

        Returns:
            A list of JobHandle objects representing the jobs.
        """
        params = {"jobTemplateName": job_template_name} if job_template_name else {}
        url = self.API_METALAKES_JOB_RUNS_PATH.format(encode_string(self.name()))
        response = self.rest_client.get(
            url, params=params, error_handler=JOB_ERROR_HANDLER
        )
        resp = JobListResponse.from_json(response.body, infer_missing=True)
        resp.validate()

        return [GenericJobHandle(dto) for dto in resp.jobs()]

    def get_job(self, job_id: str) -> JobHandle:
        """Retrieves a job by its ID.

        Args:
            job_id: The ID of the job to retrieve.

        Returns:
            The JobHandle representing the job if found, otherwise raises an exception.

        Raises:
            NoSuchJobException: If no job with the specified ID exists.
        """
        if not job_id or not job_id.strip():
            raise ValueError("Job ID cannot be null or empty")

        url = (
            f"{self.API_METALAKES_JOB_RUNS_PATH.format(encode_string(self.name()))}"
            f"/{encode_string(job_id)}"
        )
        response = self.rest_client.get(url, error_handler=JOB_ERROR_HANDLER)
        resp = JobResponse.from_json(response.body, infer_missing=True)
        resp.validate()

        return GenericJobHandle(resp.job())

    def run_job(self, job_template_name: str, job_conf: Dict[str, str]) -> JobHandle:
        """Runs a job based on the specified job template and configuration.

        Args:
            job_template_name: The name of the job template to use for running the job.
            job_conf: A dictionary containing the configuration for the job.

        Returns:
            A JobHandle representing the started job.

        Raises:
            NoSuchJobTemplateException: If no job template with the specified name exists.
        """
        url = f"{self.API_METALAKES_JOB_RUNS_PATH.format(encode_string(self.name()))}"
        request = JobRunRequest(job_template_name, job_conf)

        response = self.rest_client.post(
            url, json=request, error_handler=JOB_ERROR_HANDLER
        )
        resp = JobResponse.from_json(response.body, infer_missing=True)
        resp.validate()

        return GenericJobHandle(resp.job())

    def cancel_job(self, job_id: str) -> JobHandle:
        """Cancels a job by its ID.

        Args:
            job_id: The ID of the job to cancel.

        Returns:
            A JobHandle representing the cancelled job.

        Raises:
            NoSuchJobException: If no job with the specified ID exists.
        """
        if not job_id or not job_id.strip():
            raise ValueError("Job ID cannot be null or empty")

        url = (
            f"{self.API_METALAKES_JOB_RUNS_PATH.format(encode_string(self.name()))}"
            f"/{encode_string(job_id)}"
        )
        response = self.rest_client.post(url, error_handler=JOB_ERROR_HANDLER)
        resp = JobResponse.from_json(response.body, infer_missing=True)
        resp.validate()

        return GenericJobHandle(resp.job())

    #########
    # Tag operations
    #########
    def list_tags(self) -> list[str]:
        """List all the tag names under a metalake.

        Returns:
            list[str]: The list of tag names.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
        """
        url = self.API_METALAKES_TAGS_PATH.format(encode_string(self.name()))

        response = self.rest_client.get(url, error_handler=TAG_ERROR_HANDLER)
        resp = TagNamesListResponse.from_json(response.body, infer_missing=True)
        resp.validate()

        return resp.tag_names()

    def list_tags_info(self) -> List[Tag]:
        """
        List tags information under a metalake.

        Returns:
            list[Tag]: The list of tag information.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
        """
        # TODO implement list_tags_info
        raise NotImplementedError()

    def get_tag(self, tag_name) -> Tag:
        """
        Get a tag by its name under a metalake.

        Args:
            tag_name (str): The name of the tag.

        Returns:
            Tag: The tag information.

        Raises:
            NoSuchTagException: If the tag does not exist.
        """
        Precondition.check_string_not_empty(
            tag_name, "tag name must not be null or empty"
        )
        url = self.API_METALAKES_TAG_PATH.format(
            encode_string(self.name()), encode_string(tag_name)
        )
        response = self.rest_client.get(url, error_handler=TAG_ERROR_HANDLER)

        tag_resp = TagResponse.from_json(response.body, infer_missing=True)
        tag_resp.validate()

        return GenericTag(self.name(), tag_resp.tag(), self.rest_client)

    def create_tag(self, tag_name, comment, properties) -> Tag:
        """
        Create a new tag under a metalake.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
            TagAlreadyExistsException: If the tag already exists.

        Args:
            tag_name (str): The name of the tag.
            comment (str): The comment of the tag.
            properties (dict[str, str]): The properties of the tag.

        Returns:
            Tag: The tag information.
        """
        tag_create_request = TagCreateRequest(
            tag_name,
            comment,
            properties,
        )
        tag_create_request.validate()

        url = self.API_METALAKES_TAGS_PATH.format(encode_string(self.name()))

        response = self.rest_client.post(
            url,
            json=tag_create_request,
            error_handler=TAG_ERROR_HANDLER,
        )
        tag_resp = TagResponse.from_json(response.body, infer_missing=True)
        tag_resp.validate()

        return GenericTag(self.name(), tag_resp.tag(), self.rest_client)

    def alter_tag(self, tag_name, *changes) -> Tag:
        """
        Alter a tag under a metalake.

        Args:
            tag_name (str): The name of the tag.
            changes (TagChange): The changes to apply to the tag.

        Returns:
            Tag: The altered tag.

        Raises:
            NoSuchTagException: If the tag does not exist.
            NoSuchMetalakeException: If the metalake does not exist.
        """
        # TODO implement alter_tag
        raise NotImplementedError()

    def delete_tag(self, tag_name) -> bool:
        """
        Delete a tag under a metalake.

        Args:
            tag_name (str): The name of the tag.

        Returns:
            bool: True if the tag was deleted, False otherwise.

        Raises:
            NoSuchTagException: If the tag does not exist.
            NoSuchMetalakeException: If the metalake does not exist.
        """
        # TODO implement delete_tag
        raise NotImplementedError()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from dataclasses import dataclass, field
from typing import Optional, List, Dict

from dataclasses_json import config

from gravitino.api.job.job_template_change import TemplateUpdate, SparkTemplateUpdate
from gravitino.dto.job.template_update_dto import TemplateUpdateDTO


@dataclass
class SparkTemplateUpdateDTO(TemplateUpdateDTO):
    """DTO for updating Spark job templates."""

    _new_class_name: Optional[str] = field(metadata=config(field_name="newClassName"))
    _new_jars: Optional[List[str]] = field(metadata=config(field_name="newJars"))
    _new_files: Optional[List[str]] = field(metadata=config(field_name="newFiles"))
    _new_archives: Optional[List[str]] = field(
        metadata=config(field_name="newArchives")
    )
    _new_configs: Optional[Dict[str, str]] = field(
        metadata=config(field_name="newConfigs")
    )

    def __init__(
        self,
        new_executable: Optional[str] = None,
        new_arguments: Optional[List[str]] = None,
        new_environments: Optional[dict] = None,
        new_custom_fields: Optional[dict] = None,
        new_class_name: Optional[str] = None,
        new_jars: Optional[List[str]] = None,
        new_files: Optional[List[str]] = None,
        new_archives: Optional[List[str]] = None,
        new_configs: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            _type="spark",
            _new_executable=new_executable,
            _new_arguments=new_arguments,
            _new_environments=new_environments,
            _new_custom_fields=new_custom_fields,
        )
        self._new_class_name = new_class_name
        self._new_jars = new_jars
        self._new_files = new_files
        self._new_archives = new_archives
        self._new_configs = new_configs

    def to_template_update(self) -> TemplateUpdate:
        return SparkTemplateUpdate(
            new_executable=self._new_executable,
            new_arguments=self._new_arguments,
            new_environments=self._new_environments,
            new_custom_fields=self._new_custom_fields,
            new_class_name=self._new_class_name,
            new_jars=self._new_jars,
            new_files=self._new_files,
            new_archives=self._new_archives,
            new_configs=self._new_configs,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import config

from gravitino.api.catalog import Catalog
from gravitino.rest.rest_message import RESTRequest


@dataclass
class CatalogCreateRequest(RESTRequest):
    """Represents a request to create a catalog."""

    _name: str = field(metadata=config(field_name="name"))
    _type: Catalog.Type = field(
        metadata=config(
            field_name="type",
            encoder=Catalog.Type.type_serialize,
            decoder=Catalog.Type.type_deserialize,
        )
    )
    _provider: Optional[str] = field(metadata=config(field_name="provider"))
    _comment: Optional[str] = field(metadata=config(field_name="comment"))
    _properties: Optional[Dict[str, str]] = field(
        metadata=config(field_name="properties")
    )

    def __init__(
        self,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
    ):
        self._name = name
        self._type = catalog_type
        self._provider = provider
        self._comment = comment
        self._properties = properties

    def validate(self):
        """Validates the fields of the request.

        Raises:
            IllegalArgumentException if name or type are not set.
        """
        if not self._name:
            raise ValueError('"name" field is required and cannot be empty')
        if not self._type:
            raise ValueError('"catalog_type" field is required and cannot be empty')
        if not self._provider and not self._type.supports_managed_catalog:
            raise ValueError(
                '"provider" field is required and cannot be empty for catalog type '
                "that does not support managed catalog"
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from typing import List, Optional

from gravitino.api.auditable import Auditable
from gravitino.api.function.function_definition import FunctionDefinition
from gravitino.api.function.function_type import FunctionType


class Function(Auditable):
    """An interface representing a user-defined function under a schema Namespace.

    A function is a reusable computational unit that can be invoked within queries across
    different compute engines. Users can register a function in Gravitino to manage the
    function metadata and enable cross-engine function sharing. The typical use case is
    to define custom business logic once and reuse it across multiple compute engines
    like Spark, Trino, and AI engines.

    A function is characterized by its name, type (scalar for row-by-row operations,
    aggregate for group operations, or table-valued for set-returning operations),
    whether it is deterministic, and its definitions that contain parameters, return
    type or columns (for table function), and implementations for different runtime engines.
    """

    @abstractmethod
    def name(self) -> str:
        """Returns the function name."""
        pass

    @abstractmethod
    def function_type(self) -> FunctionType:
        """Returns the function type."""
        pass

    @abstractmethod
    def deterministic(self) -> bool:
        """Returns whether the function is deterministic."""
        pass

    def comment(self) -> Optional[str]:
        """Returns the optional comment of the function."""
        return None

    @abstractmethod
    def definitions(self) -> List[FunctionDefinition]:
        """Returns the definitions of the function."""
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import functools
import importlib
import logging
import re
from typing import Dict, Optional, Callable

import fsspec

from gravitino.audit.fileset_data_operation import FilesetDataOperation
from gravitino.exceptions.base import (
    GravitinoRuntimeException,
    NoSuchCatalogException,
    CatalogNotInUseException,
    NoSuchFilesetException,
    NoSuchLocationNameException,
)
from gravitino.filesystem.gvfs_base_operations import (
    BaseGVFSOperations,
)
from gravitino.filesystem.gvfs_config import GVFSConfig
from gravitino.filesystem.gvfs_default_operations import DefaultGVFSOperations
from gravitino.filesystem.gvfs_hook import DEFAULT_HOOK, GravitinoVirtualFileSystemHook

logger = logging.getLogger(__name__)

PROTOCOL_NAME = "gvfs"


class GravitinoVirtualFileSystem(fsspec.AbstractFileSystem):
    """This is a virtual file system that users can access `fileset` and
    other resources.

    It obtains the actual storage location corresponding to the resource from the
    Gravitino server, and creates an independent file system for it to act as an agent for users to
    access the underlying storage.
    """

    # Override the parent variable
    protocol = PROTOCOL_NAME
    _identifier_pattern = re.compile("^fileset/([^/]+)/([^/]+)/([^/]+)(?:/[^/]+)*/?$")
    SLASH = "/"

    def __init__(
        self,
        server_uri: str = None,
        metalake_name: str = None,
        options: Dict = None,
        **kwargs,
    ):
        """Initialize the GravitinoVirtualFileSystem.
        :param server_uri: Gravitino server URI
        :param metalake_name: Gravitino metalake name
        :param options: Options for the GravitinoVirtualFileSystem
        :param kwargs: Extra args for super filesystem
        """
        self._kwargs = dict(kwargs)
        self._hook = self._get_hook_class(options)
        self._hook.initialize(options)
        self._operations = self._get_gvfs_operations_class(
            server_uri, metalake_name, options
        )
        self._hook.set_operations_context(self._operations)

        super().__init__(**kwargs)

    @property
    def fsid(self):
        return PROTOCOL_NAME

    @property
    def hook(self):
        return self._hook

    @property
    def operations(self):
        return self._operations

    @staticmethod
    def _with_exception_translation(operation: FilesetDataOperation):
        """
        Decorator to translate fileset metadata not found exceptions into FileNotFoundError.
        :param operation: The operation being performed.
        """

        def decorator(func: Callable):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except (NoSuchCatalogException, CatalogNotInUseException) as e:
                    message = f"Cannot get fileset catalog during {operation}"
                    logger.warning("%s, %s", message, str(e))
                    raise FileNotFoundError(message) from e
                except NoSuchFilesetException as e:
                    message = f"Cannot get fileset during {operation}"
                    logger.warning("%s, %s", message, str(e))
                    raise FileNotFoundError(message) from e
                except NoSuchLocationNameException as e:
                    message = f"Cannot find location name during {operation}"
                    logger.warning("%s, %s", message, str(e))
                    raise FileNotFoundError(message) from e

            return wrapper

        return decorator

    def sign(self, path, expiration=None, **kwargs):
        """We do not support to create a signed URL representing the given path in gvfs."""
        raise GravitinoRuntimeException(
            "Sign is not implemented for Gravitino Virtual FileSystem."
        )

    def ls(self, path, detail=True, **kwargs):
        """List the files and directories info of the path.
        :param path: Virtual fileset path
        :param detail: Whether to show the details for the files and directories info
        :param kwargs: Extra args
        :return If details is true, returns a list of file info dicts, else returns a list of file paths
        """
        try:
            new_path = self._hook.pre_ls(path, detail, **kwargs)
            decorated_ls = self._with_exception_translation(
                FilesetDataOperation.LIST_STATUS
            )(self._operations.ls)
            result = decorated_ls(new_path, detail, **kwargs)
            return self._hook.post_ls(detail, result, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Hook will either provide fallback value or re-raise the exception
            return self._hook.on_ls_failure(path, e, **kwargs)

    def info(self, path, **kwargs):
        """Get file info.
        :param path: Virtual fileset path
        :param kwargs: Extra args
        :return A file info dict
        """
        try:
            new_path = self._hook.pre_info(path, **kwargs)
            decorated_info = self._with_exception_translation(
                FilesetDataOperation.GET_FILE_STATUS
            )(self._operations.info)
            result = decorated_info(new_path, **kwargs)
            return self._hook.post_info(
                result,
                **kwargs,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Hook will either provide fallback value or re-raise the exception
            return self._hook.on_info_failure(path, e, **kwargs)

    def exists(self, path, **kwargs):
        """Check if a file or a directory exists.
        :param path: Virtual fileset path
        :param kwargs: Extra args
        :return If a file or directory exists, it returns True, otherwise False
        """
        try:
            new_path = self._hook.pre_exists(path, **kwargs)
            decorated_exists = self._with_exception_translation(
                FilesetDataOperation.EXISTS
            )(self._operations.exists)
            try:
                result = decorated_exists(new_path, **kwargs)
            except FileNotFoundError:
                return False

            return self._hook.post_exists(
                new_path,
                result,
                **kwargs,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Hook will either provide fallback value or re-raise the exception
            return self._hook.on_exists_failure(path, e, **kwargs)

    def cp_file(self, path1, path2, **kwargs):
        """Copy a file.
        :param path1: Virtual src fileset path
        :param path2: Virtual dst fileset path, should be consistent with the src path fileset identifier
        :param kwargs: Extra args
        """
        try:
            new_path1, new_path2 = self._hook.pre_cp_file(path1, path2, **kwargs)
            decorated_cp_file = self._with_exception_translation(
                FilesetDataOperation.COPY_FILE
            )(self._operations.cp_file)
            decorated_cp_file(new_path1, new_path2, **kwargs)
            self._hook.post_cp_file(new_path1, new_path2, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_cp_file_failure(path1, path2, e, **kwargs)

    def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
        """Move a file to another directory.
         This can move a file to another existing directory.
         If the target path directory does not exist, an exception will be thrown.
        :param path1: Virtual src fileset path
        :param path2: Virtual dst fileset path, should be consistent with the src path fileset identifier
        :param recursive: Whether to move recursively
        :param maxdepth: Maximum depth of recursive move
        :param kwargs: Extra args
        """
        try:
            new_path1, new_path2 = self._hook.pre_mv(
                path1, path2, recursive, maxdepth, **kwargs
            )
            decorated_mv = self._with_exception_translation(
                FilesetDataOperation.RENAME
            )(self._operations.mv)
            decorated_mv(new_path1, new_path2, recursive, maxdepth, **kwargs)
            self._hook.post_mv(new_path1, new_path2, recursive, maxdepth, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_mv_failure(path1, path2, recursive, maxdepth, e, **kwargs)

    def _rm(self, path):
        raise GravitinoRuntimeException(
            "Deprecated method, use `rm_file` method instead."
        )

    def rm(self, path, recursive=False, maxdepth=None):
        """Remove a file or directory.
        :param path: Virtual fileset path
        :param recursive: Whether to remove the directory recursively.
                When removing a directory, this parameter should be True.
        :param maxdepth: The maximum depth to remove the directory recursively.
        """
        try:
            new_path = self._hook.pre_rm(path, recursive, maxdepth)
            decorated_rm = self._with_exception_translation(
                FilesetDataOperation.DELETE
            )(self._operations.rm)
            decorated_rm(new_path, recursive, maxdepth)
            self._hook.post_rm(new_path, recursive, maxdepth)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_rm_failure(path, recursive, maxdepth, e)

    def rm_file(self, path):
        """Remove a file.
        :param path: Virtual fileset path
        """
        try:
            new_path = self._hook.pre_rm_file(path)
            decorated_rm_file = self._with_exception_translation(
                FilesetDataOperation.DELETE
            )(self._operations.rm_file)
            decorated_rm_file(new_path)
            self._hook.post_rm_file(new_path)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_rm_file_failure(path, e)

    def rmdir(self, path):
        """Remove a directory.
        It will delete a directory and all its contents recursively for PyArrow.HadoopFileSystem.
        And it will throw an exception if delete a directory which is non-empty for LocalFileSystem.
        :param path: Virtual fileset path
        """
        try:
            new_path = self._hook.pre_rmdir(path)
            decorated_rmdir = self._with_exception_translation(
                FilesetDataOperation.DELETE
            )(self._operations.rmdir)
            decorated_rmdir(new_path)
            self._hook.post_rmdir(new_path)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_rmdir_failure(path, e)

    def open(
        self,
        path,
        mode="rb",
        block_size=None,
        cache_options=None,
        compression=None,
        **kwargs,
    ):
        """Open a file to read/write/append.
        :param path: Virtual fileset path
        :param mode: The mode now supports: rb(read), wb(write), ab(append). See builtin ``open()``
        :param block_size: Some indication of buffering - this is a value in bytes
        :param cache_options: Extra arguments to pass through to the cache
        :param compression: If given, open file using compression codec
        :param kwargs: Extra args
        :return A file-like object from the filesystem
        """
        try:
            new_path = self._hook.pre_open(
                path, mode, block_size, cache_options, compression, **kwargs
            )
            if mode in ("w", "wb"):
                data_operation = FilesetDataOperation.OPEN_AND_WRITE
            elif mode in ("a", "ab"):
                data_operation = FilesetDataOperation.OPEN_AND_APPEND
            else:
                data_operation = FilesetDataOperation.OPEN

            decorated_open = self._with_exception_translation(data_operation)(
                self._operations.open
            )
            try:
                result = decorated_open(
                    new_path,
                    mode,
                    block_size,
                    cache_options,
                    compression,
                    **kwargs,
                )
            except FileNotFoundError as e:
                if mode in ("w", "wb", "x", "xb", "a", "ab"):
                    raise OSError(
                        f"Fileset is not found for path: {new_path} for operation OPEN. This "
                        f"may be caused by fileset related metadata not found or not in use "
                        f"in Gravitino,"
                    ) from e
                raise
            return self._hook.post_open(
                new_path,
                mode,
                block_size,
                cache_options,
                compression,
                result,
                **kwargs,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Hook will either provide fallback value or re-raise the exception
            return self._hook.on_open_failure(
                path, mode, block_size, cache_options, compression, e, **kwargs
            )

    def mkdir(self, path, create_parents=True, **kwargs):
        """Make a directory.
        if create_parents=True, this is equivalent to ``makedirs``.

        :param path: Virtual fileset path
        :param create_parents: Create parent directories if missing when set to True
        :param kwargs: Extra args
        """
        try:
            new_path = self._hook.pre_mkdir(path, create_parents, **kwargs)
            try:
                self._operations.mkdir(new_path, create_parents, **kwargs)
            except (
                NoSuchCatalogException,
                CatalogNotInUseException,
                NoSuchFilesetException,
                NoSuchLocationNameException,
            ) as e:
                raise OSError(
                    f"Fileset is not found for path: {new_path} for operation MKDIRS. This "
                    f"may be caused by fileset related metadata not found or not in use "
                    f"in Gravitino,"
                ) from e
            self._hook.post_mkdir(new_path, create_parents, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_mkdir_failure(path, create_parents, e, **kwargs)

    def makedirs(self, path, exist_ok=True):
        """Make a directory recursively.
        :param path: Virtual fileset path
        :param exist_ok: Continue if a directory already exists
        """
        try:
            new_path = self._hook.pre_makedirs(path, exist_ok)
            try:
                self._operations.makedirs(new_path, exist_ok)
            except (
                NoSuchCatalogException,
                CatalogNotInUseException,
                NoSuchFilesetException,
                NoSuchLocationNameException,
            ) as e:
                raise OSError(
                    f"Fileset is not found for path: {new_path} for operation MKDIRS. This "
                    f"may be caused by fileset related metadata not found or not in use "
                    f"in Gravitino,"
                ) from e
            self._hook.post_makedirs(new_path, exist_ok)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_makedirs_failure(path, exist_ok, e)

    def created(self, path):
        """Return the created timestamp of a file as a datetime.datetime
        Only supports for `fsspec.LocalFileSystem` now.
        :param path: Virtual fileset path
        :return Created time(datetime.datetime)
        """
        try:
            new_path = self._hook.pre_created(path)
            decorated_created = self._with_exception_translation(
                FilesetDataOperation.CREATED_TIME
            )(self._operations.created)
            result = decorated_created(new_path)
            return self._hook.post_created(
                new_path,
                result,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Hook will either provide fallback value or re-raise the exception
            return self._hook.on_created_failure(path, e)

    def modified(self, path):
        """Returns the modified time of the path file if it exists.
        :param path: Virtual fileset path
        :return Modified time(datetime.datetime)
        """
        try:
            new_path = self._hook.pre_modified(path)
            decorated_modified = self._with_exception_translation(
                FilesetDataOperation.MODIFIED_TIME
            )(self._operations.modified)
            result = decorated_modified(new_path)
            return self._hook.post_modified(
                new_path,
                result,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Hook will either provide fallback value or re-raise the exception
            return self._hook.on_modified_failure(path, e)

    def cat_file(self, path, start=None, end=None, **kwargs):
        """Get the content of a file.
        :param path: Virtual fileset path
        :param start: The offset in bytes to start reading from. It can be None.
        :param end: The offset in bytes to end reading at. It can be None.
        :param kwargs: Extra args
        :return File content
        """
        try:
            new_path = self._hook.pre_cat_file(path, start, end, **kwargs)
            decorated_cat_file = self._with_exception_translation(
                FilesetDataOperation.CAT_FILE
            )(self._operations.cat_file)
            result = decorated_cat_file(new_path, start, end, **kwargs)
            return self._hook.post_cat_file(
                new_path,
                start,
                end,
                result,
                **kwargs,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Hook will either provide fallback value or re-raise the exception
            return self._hook.on_cat_file_failure(path, start, end, e, **kwargs)

    def get_file(self, rpath, lpath, callback=None, outfile=None, **kwargs):
        """Copy single remote file to local.
        :param rpath: Remote file path
        :param lpath: Local file path
        :param callback: The callback class
        :param outfile: The output file path
        :param kwargs: Extra args
        """
        try:
            new_rpath = self._hook.pre_get_file(
                rpath, lpath, callback, outfile, **kwargs
            )
            decorated_get_file = self._with_exception_translation(
                FilesetDataOperation.GET_FILE_STATUS
            )(self._operations.get_file)
            decorated_get_file(
                new_rpath,
                lpath,
                callback,
                outfile,
                **kwargs,
            )
            self._hook.post_get_file(new_rpath, lpath, outfile, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._hook.on_get_file_failure(rpath, lpath, callback, outfile, e, **kwargs)

    def _get_hook_class(
        self, options: Optional[Dict[str, str]]
    ) -> GravitinoVirtualFileSystemHook:
        hook_class = (
            None if options is None else options.get(GVFSConfig.GVFS_FILESYSTEM_HOOK)
        )

        if hook_class is not None:
            module_name, class_name = hook_class.rsplit(".", 1)
            module = importlib.import_module(module_name)
            loaded_class = getattr(module, class_name)
            return loaded_class()

        return DEFAULT_HOOK

    def _get_gvfs_operations_class(
        self,
        server_uri: str = None,
        metalake_name: str = None,
        options: Dict = None,
    ) -> BaseGVFSOperations:
        operations_class = (
            None
            if options is None
            else options.get(GVFSConfig.GVFS_FILESYSTEM_OPERATIONS)
        )
        if operations_class is not None:
            module_name, class_name = operations_class.rsplit(".", 1)
            module = importlib.import_module(module_name)
            loaded_class = getattr(module, class_name)
            return loaded_class(server_uri, metalake_name, options)

        return DefaultGVFSOperations(
            server_uri=server_uri,
            metalake_name=metalake_name,
            options=options,
            **self._kwargs,
        )


fsspec.register_implementation(PROTOCOL_NAME, GravitinoVirtualFileSystem)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.function.function import Function
from gravitino.api.function.function_definition import FunctionDefinition
from gravitino.api.function.function_type import FunctionType
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.function.function_definition_dto import FunctionDefinitionDTO


def _encode_function_type(func_type: FunctionType) -> str:
    """Encode FunctionType to string."""
    if func_type is None:
        return None
    return func_type.value


def _decode_function_type(func_type_str: str) -> FunctionType:
    """Decode string to FunctionType."""
    return FunctionType.from_string(func_type_str)


@dataclass
class FunctionDTO(Function, DataClassJsonMixin):
    """Represents a Function DTO (Data Transfer Object)."""

    _name: str = field(metadata=config(field_name="name"))
    _function_type: FunctionType = field(
        metadata=config(
            field_name="functionType",
            encoder=_encode_function_type,
            decoder=_decode_function_type,
        )
    )
    _deterministic: bool = field(metadata=config(field_name="deterministic"))
    _comment: Optional[str] = field(default=None, metadata=config(field_name="comment"))
    _definitions: Optional[List[FunctionDefinitionDTO]] = field(
        default=None, metadata=config(field_name="definitions")
    )
    _audit: Optional[AuditDTO] = field(
        default=None, metadata=config(field_name="audit")
    )

    def __post_init__(self):
        if self._function_type is None:
            raise ValueError("Function type cannot be null or empty")

    def name(self) -> str:
        """Returns the function name."""
        return self._name

    def function_type(self) -> FunctionType:
        """Returns the function type."""
        return self._function_type

    def deterministic(self) -> bool:
        """Returns whether the function is deterministic."""
        return self._deterministic

    def comment(self) -> Optional[str]:
        """Returns the optional comment of the function."""
        return self._comment

    def definitions(self) -> Optional[List[FunctionDefinition]]:
        """Returns the definitions of the function."""
        if self._definitions is None:
            return None
        return list(self._definitions)

    def audit_info(self) -> Optional[AuditDTO]:
        """Returns the audit information."""
        return self._audit

    def __eq__(self, other) -> bool:
        if not isinstance(other, FunctionDTO):
            return False
        return (
            self._name == other._name
            and self._function_type == other._function_type
            and self._deterministic == other._deterministic
            and self._comment == other._comment
            and self._definitions == other._definitions
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._name,
                self._function_type,
                self._deterministic,
                self._comment,
            )
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import ClassVar, List, Optional, final

from gravitino.api.rel.indexes.index import Index


class Indexes:
    """Helper methods to create index to pass into Apache Gravitino.

    Attributes:
        EMPTY_INDEXES (List[Index]):
            An empty array of indexes.
        DEFAULT_MYSQL_PRIMARY_KEY_NAME (str):
            MySQL does not support setting the name of the primary key,
            so the default name is used.
    """

    EMPTY_INDEXES: ClassVar[List[Index]] = []
    DEFAULT_MYSQL_PRIMARY_KEY_NAME: ClassVar[str] = "PRIMARY"

    @staticmethod
    def of(
        index_type: Index.IndexType, name: Optional[str], field_names: List[List[str]]
    ) -> Index:
        return Indexes.IndexImpl(index_type, name, field_names)

    @staticmethod
    def unique(name: str, field_names: List[List[str]]) -> Index:
        return Indexes.IndexImpl(Index.IndexType.UNIQUE_KEY, name, field_names)

    @staticmethod
    def primary(name: str, field_names: List[List[str]]) -> Index:
        return Indexes.IndexImpl(Index.IndexType.PRIMARY_KEY, name, field_names)

    @staticmethod
    def create_mysql_primary_key(field_names: List[List[str]]) -> Index:
        return Indexes.primary(Indexes.DEFAULT_MYSQL_PRIMARY_KEY_NAME, field_names)

    @final
    class IndexImpl(Index):
        def __init__(
            self,
            index_type: Index.IndexType,
            name: Optional[str],
            field_names: List[List[str]],
        ):
            self._index_type = index_type
            self._name = name
            self._field_names = field_names

        def type(self) -> Index.IndexType:
            return self._index_type

        def name(self) -> Optional[str]:
            return self._name

        def field_names(self) -> List[List[str]]:
            return self._field_names



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from typing import List
from gravitino.api.credential.supports_credentials import SupportsCredentials
from gravitino.api.credential.credential import Credential
from gravitino.api.metadata_object import MetadataObject
from gravitino.audit.caller_context import CallerContext, CallerContextHolder
from gravitino.dto.credential_dto import CredentialDTO
from gravitino.dto.responses.credential_response import CredentialResponse
from gravitino.exceptions.handlers.credential_error_handler import (
    CREDENTIAL_ERROR_HANDLER,
)
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient
from gravitino.utils.credential_factory import CredentialFactory

logger = logging.getLogger(__name__)


class MetadataObjectCredentialOperations(SupportsCredentials):
    _rest_client: HTTPClient
    """The REST client to communicate with the REST server"""

    _request_path: str
    """The REST API path to do credential operations"""

    def __init__(
        self,
        metalake_name: str,
        metadata_object: MetadataObject,
        rest_client: HTTPClient,
    ):
        self._rest_client = rest_client
        metadata_object_type = metadata_object.type().value
        metadata_object_fullname = metadata_object.full_name()
        self._request_path = (
            f"api/metalakes/{encode_string(metalake_name)}/objects/{metadata_object_type}/"
            f"{encode_string(metadata_object_fullname)}/credentials"
        )

    def get_credentials(self) -> List[Credential]:
        try:
            caller_context: CallerContext = CallerContextHolder.get()
            resp = self._rest_client.get(
                self._request_path,
                headers=(
                    caller_context.context() if caller_context is not None else None
                ),
                error_handler=CREDENTIAL_ERROR_HANDLER,
            )
        finally:
            CallerContextHolder.remove()

        credential_resp = CredentialResponse.from_json(resp.body, infer_missing=True)
        credential_resp.validate()
        credential_dtos = credential_resp.credentials()
        return self.to_credentials(credential_dtos)

    def to_credentials(self, credentials: List[CredentialDTO]) -> List[Credential]:
        return [self.to_credential(credential) for credential in credentials]

    def to_credential(self, credential_dto: CredentialDTO) -> Credential:
        return CredentialFactory.create(
            credential_dto.credential_type(),
            credential_dto.credential_info(),
            credential_dto.expire_time_in_ms(),
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Optional, Dict, List

from gravitino.api.model.model_version import ModelVersion
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.model_version_dto import ModelVersionDTO


class GenericModelVersion(ModelVersion):
    _model_version_dto: ModelVersionDTO
    """The model version DTO object."""

    def __init__(self, model_version_dto: ModelVersionDTO):
        self._model_version_dto = model_version_dto

    def version(self) -> int:
        return self._model_version_dto.version()

    def comment(self) -> Optional[str]:
        return self._model_version_dto.comment()

    def aliases(self) -> List[str]:
        return self._model_version_dto.aliases()

    def uris(self) -> Dict[str, str]:
        return self._model_version_dto.uris()

    def properties(self) -> Dict[str, str]:
        return self._model_version_dto.properties()

    def audit_info(self) -> AuditDTO:
        return self._model_version_dto.audit_info()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Dict, Optional

from gravitino.api.function.function_impl import FunctionImpl
from gravitino.api.function.function_resources import FunctionResources
from gravitino.utils.precondition import Precondition


class SQLImpl(FunctionImpl):
    """SQL implementation with runtime and SQL body."""

    def __init__(
        self,
        runtime: FunctionImpl.RuntimeType,
        sql: str,
        resources: Optional[FunctionResources] = None,
        properties: Optional[Dict[str, str]] = None,
    ):
        """Create a SQLImpl instance.

        Args:
            runtime: The runtime of the function implementation.
            sql: The SQL that defines the function.
            resources: The resources required by the function implementation.
            properties: The properties of the function implementation.

        Raises:
            ValueError: If sql is null or empty.
        """
        super().__init__(FunctionImpl.Language.SQL, runtime, resources, properties)
        if not sql or not sql.strip():
            raise ValueError("SQL text cannot be null or empty")
        self._sql = sql

    def sql(self) -> str:
        """Returns the SQL that defines the function."""
        return self._sql

    @staticmethod
    def builder() -> "SQLImpl.Builder":
        """Returns a new Builder for SQLImpl."""
        return SQLImpl.Builder()

    def __eq__(self, other) -> bool:
        if not isinstance(other, SQLImpl):
            return False
        return (
            self.language() == other.language()
            and self.runtime() == other.runtime()
            and self.resources() == other.resources()
            and self.properties() == other.properties()
            and self._sql == other._sql
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.language(),
                self.runtime(),
                self.resources(),
                tuple(sorted(self.properties().items())),
                self._sql,
            )
        )

    def __repr__(self) -> str:
        return (
            f"SQLImpl(language={self.language()}, runtime={self.runtime()}, "
            f"sql='{self._sql}', resources={self.resources()}, "
            f"properties={self.properties()})"
        )

    class Builder:
        """Builder for SQLImpl."""

        def __init__(self):
            self._runtime: Optional[FunctionImpl.RuntimeType] = None
            self._sql: Optional[str] = None
            self._resources: Optional[FunctionResources] = None
            self._properties: Optional[Dict[str, str]] = None

        def with_runtime_type(
            self, runtime: FunctionImpl.RuntimeType
        ) -> "SQLImpl.Builder":
            """Sets the runtime type.

            Args:
                runtime: The runtime of the function implementation.

            Returns:
                The builder instance.
            """
            self._runtime = runtime
            return self

        def with_sql(self, sql: str) -> "SQLImpl.Builder":
            """Sets the SQL text.

            Args:
                sql: The SQL that defines the function.

            Returns:
                The builder instance.
            """
            self._sql = sql
            return self

        def with_resources(self, resources: FunctionResources) -> "SQLImpl.Builder":
            """Sets the resources.

            Args:
                resources: The resources required by the function implementation.

            Returns:
                The builder instance.
            """
            self._resources = resources
            return self

        def with_properties(self, properties: Dict[str, str]) -> "SQLImpl.Builder":
            """Sets the properties.

            Args:
                properties: The properties of the function implementation.

            Returns:
                The builder instance.
            """
            self._properties = properties
            return self

        def build(self) -> "SQLImpl":
            """Builds a SQLImpl instance.

            Returns:
                A new SQLImpl instance.

            Raises:
                IllegalArgumentException: If required fields are not set.
            """
            Precondition.check_argument(
                self._runtime is not None, "Runtime type cannot be null"
            )
            Precondition.check_argument(self._sql is not None, "SQL cannot be null")

            return SQLImpl(
                runtime=self._runtime,
                sql=self._sql,
                resources=self._resources,
                properties=self._properties,
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import List, Optional

from gravitino.api.function.function_column import FunctionColumn
from gravitino.api.function.function_impl import FunctionImpl
from gravitino.api.function.function_param import FunctionParam
from gravitino.api.rel.types.type import Type


class FunctionDefinition(ABC):
    """A function definition that pairs a specific parameter list with its implementations.

    A single function can include multiple definitions (overloads), each with distinct
    parameters, return types, and implementations.

    For scalar or aggregate functions, use return_type() to specify the return type.
    For table-valued functions, use return_columns() to specify the output columns.
    """

    EMPTY_COLUMNS: List[FunctionColumn] = []
    """An empty list of FunctionColumn."""

    @abstractmethod
    def parameters(self) -> List[FunctionParam]:
        """Returns the parameters for this definition.

        Returns:
            The parameters for this definition. May be an empty list for a no-arg definition.
        """
        pass

    def return_type(self) -> Optional[Type]:
        """The return type for scalar or aggregate function definitions.

        Returns:
            The return type, or None if this is a table-valued function definition.
        """
        return None

    def return_columns(self) -> List[FunctionColumn]:
        """The output columns for a table-valued function definition.

        A table-valued function is a function that returns a table instead of a
        scalar value or an aggregate result. The returned table has a fixed schema
        defined by the columns returned from this method.

        Returns:
            The output columns that define the schema of the table returned by
            this definition, or an empty list if this is a scalar or aggregate
            function definition.
        """
        return self.EMPTY_COLUMNS

    @abstractmethod
    def impls(self) -> List[FunctionImpl]:
        """Returns the implementations associated with this definition."""
        pass


class FunctionDefinitions:
    """Factory class for creating FunctionDefinition instances."""

    class SimpleFunctionDefinition(FunctionDefinition):
        """Simple implementation of FunctionDefinition."""

        def __init__(
            self,
            parameters: List[FunctionParam],
            return_type: Optional[Type],
            return_columns: Optional[List[FunctionColumn]],
            impls: List[FunctionImpl],
        ):
            self._parameters = list(parameters) if parameters else []
            self._return_type = return_type
            self._return_columns = (
                list(return_columns)
                if return_columns
                else FunctionDefinition.EMPTY_COLUMNS
            )
            if not impls:
                raise ValueError("Impls cannot be null or empty")
            self._impls = list(impls)

        def parameters(self) -> List[FunctionParam]:
            return list(self._parameters)

        def return_type(self) -> Optional[Type]:
            return self._return_type

        def return_columns(self) -> List[FunctionColumn]:
            return (
                list(self._return_columns)
                if self._return_columns
                else FunctionDefinition.EMPTY_COLUMNS
            )

        def impls(self) -> List[FunctionImpl]:
            return list(self._impls)

        def __eq__(self, other) -> bool:
            if not isinstance(other, FunctionDefinition):
                return False
            return (
                self._parameters == list(other.parameters())
                and self._return_type == other.return_type()
                and self._return_columns == list(other.return_columns())
                and self._impls == list(other.impls())
            )

        def __hash__(self) -> int:
            return hash(
                (
                    tuple(self._parameters),
                    self._return_type,
                    tuple(self._return_columns) if self._return_columns else None,
                    tuple(self._impls),
                )
            )

        def __repr__(self) -> str:
            return (
                f"FunctionDefinition(parameters={self._parameters}, "
                f"returnType={self._return_type}, "
                f"returnColumns={self._return_columns}, "
                f"impls={self._impls})"
            )

    @classmethod
    def of(
        cls,
        parameters: List[FunctionParam],
        return_type: Type,
        impls: List[FunctionImpl],
    ) -> FunctionDefinition:
        """Create a FunctionDefinition instance for a scalar or aggregate function.

        Args:
            parameters: The parameters for this definition, may be empty.
            return_type: The return type for this definition, must not be None.
            impls: The implementations for this definition, must not be empty.

        Returns:
            A FunctionDefinition instance.

        Raises:
            ValueError: If return_type is None.
        """
        if return_type is None:
            raise ValueError("Return type cannot be None")
        return cls.SimpleFunctionDefinition(parameters, return_type, None, impls)

    @classmethod
    def of_table(
        cls,
        parameters: List[FunctionParam],
        return_columns: List[FunctionColumn],
        impls: List[FunctionImpl],
    ) -> FunctionDefinition:
        """Create a FunctionDefinition instance for a table-valued function.

        Args:
            parameters: The parameters for this definition, may be empty.
            return_columns: The return columns for this definition, must not be empty.
            impls: The implementations for this definition, must not be empty.

        Returns:
            A FunctionDefinition instance.

        Raises:
            ValueError: If return_columns is None or empty.
        """
        if not return_columns:
            raise ValueError("Return columns cannot be None or empty")
        return cls.SimpleFunctionDefinition(parameters, None, return_columns, impls)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from contextlib import suppress
from typing import Any, Dict, List, cast

from gravitino.dto.rel.expressions.json_serdes._helper.serdes_utils import (
    SerdesUtils as ExpressionsSerdesUtils,
)
from gravitino.dto.rel.expressions.literal_dto import LiteralDTO
from gravitino.dto.rel.partitions.identity_partition_dto import IdentityPartitionDTO
from gravitino.dto.rel.partitions.list_partition_dto import ListPartitionDTO
from gravitino.dto.rel.partitions.partition_dto import PartitionDTO
from gravitino.dto.rel.partitions.range_partition_dto import RangePartitionDTO
from gravitino.exceptions.base import IllegalArgumentException
from gravitino.utils.precondition import Precondition
from gravitino.utils.serdes import SerdesUtilsBase


class SerdesUtils(SerdesUtilsBase):
    @classmethod
    def write_partition(cls, value: PartitionDTO) -> Dict[str, Any]:
        result = {
            cls.PARTITION_TYPE: value.type().value,
            cls.PARTITION_NAME: value.name(),
        }
        dto_data = {}
        dto_type = value.type()

        if dto_type not in PartitionDTO.Type:
            raise IOError(f"Unknown partition type: {value.type()}")

        if dto_type is PartitionDTO.Type.IDENTITY:
            dto = cast(IdentityPartitionDTO, value)
            dto_data[cls.FIELD_NAMES] = dto.field_names()
            dto_data[cls.IDENTITY_PARTITION_VALUES] = [
                ExpressionsSerdesUtils.write_function_arg(arg=arg)
                for arg in dto.values()
            ]

        if dto_type is PartitionDTO.Type.LIST:
            dto = cast(ListPartitionDTO, value)
            dto_data[cls.LIST_PARTITION_LISTS] = [
                [ExpressionsSerdesUtils.write_function_arg(arg=arg) for arg in args]
                for args in dto.lists()
            ]

        if dto_type is PartitionDTO.Type.RANGE:
            dto = cast(RangePartitionDTO, value)
            dto_data[cls.RANGE_PARTITION_UPPER] = (
                ExpressionsSerdesUtils.write_function_arg(arg=dto.upper())
            )
            dto_data[cls.RANGE_PARTITION_LOWER] = (
                ExpressionsSerdesUtils.write_function_arg(arg=dto.lower())
            )

        dto_data[cls.PARTITION_PROPERTIES] = value.properties()
        result.update(dto_data)
        return result

    @classmethod
    def read_partition(cls, data: Dict[str, Any]) -> PartitionDTO:
        Precondition.check_argument(
            isinstance(data, Dict) and len(data) > 0,
            f"Partition must be a valid JSON object, but found: {data}",
        )
        Precondition.check_argument(
            data.get(cls.PARTITION_TYPE) is not None,
            f"Partition must have a type field, but found: {data}",
        )
        dto_type = None
        with suppress(ValueError):
            dto_type = PartitionDTO.Type(data[cls.PARTITION_TYPE])

        if dto_type is PartitionDTO.Type.IDENTITY:
            Precondition.check_argument(
                isinstance(data.get(cls.FIELD_NAMES), List),
                f"Identity partition must have array of fieldNames, but found: {data}",
            )
            Precondition.check_argument(
                isinstance(data.get(cls.IDENTITY_PARTITION_VALUES), List),
                f"Identity partition must have array of values, but found: {data}",
            )
            return IdentityPartitionDTO(
                name=data[cls.PARTITION_NAME],
                field_names=data[cls.FIELD_NAMES],
                values=[
                    cast(
                        LiteralDTO, ExpressionsSerdesUtils.read_function_arg(data=value)
                    )
                    for value in data[cls.IDENTITY_PARTITION_VALUES]
                ],
                properties=data.get(cls.PARTITION_PROPERTIES, {}),
            )

        if dto_type is PartitionDTO.Type.LIST:
            Precondition.check_argument(
                cls.PARTITION_NAME in data,
                f"List partition must have name, but found: {data}",
            )
            Precondition.check_argument(
                isinstance(data.get(cls.LIST_PARTITION_LISTS), List),
                f"List partition must have array of lists, but found: {data}",
            )
            return ListPartitionDTO(
                name=data[cls.PARTITION_NAME],
                lists=[
                    [
                        cast(
                            LiteralDTO,
                            ExpressionsSerdesUtils.read_function_arg(data=value),
                        )
                        for value in values
                    ]
                    for values in data[cls.LIST_PARTITION_LISTS]
                ],
                properties=data.get(cls.PARTITION_PROPERTIES, {}),
            )

        if dto_type is PartitionDTO.Type.RANGE:
            Precondition.check_argument(
                cls.PARTITION_NAME in data,
                f"Range partition must have name, but found: {data}",
            )
            Precondition.check_argument(
                cls.RANGE_PARTITION_UPPER in data,
                f"Range partition must have upper, but found: {data}",
            )
            Precondition.check_argument(
                cls.RANGE_PARTITION_LOWER in data,
                f"Range partition must have lower, but found: {data}",
            )
            upper = cast(
                LiteralDTO,
                ExpressionsSerdesUtils.read_function_arg(
                    data[cls.RANGE_PARTITION_UPPER]
                ),
            )
            lower = cast(
                LiteralDTO,
                ExpressionsSerdesUtils.read_function_arg(
                    data[cls.RANGE_PARTITION_LOWER]
                ),
            )
            return RangePartitionDTO(
                name=data[cls.PARTITION_NAME],
                upper=upper,
                lower=lower,
                properties=data.get(cls.PARTITION_PROPERTIES, {}),
            )

        raise IllegalArgumentException(
            f"Unknown partition type: {data[cls.PARTITION_TYPE]}"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import config

from gravitino.dto.job.job_template_dto import JobTemplateDTO
from gravitino.rest.rest_message import RESTRequest


@dataclass
class JobTemplateRegisterRequest(RESTRequest):
    """Represents a request to register a job template."""

    _job_template: JobTemplateDTO = field(metadata=config(field_name="jobTemplate"))

    def job_template(self) -> Optional[JobTemplateDTO]:
        """Returns the job template to register."""
        return self._job_template

    def validate(self):
        """Validates the request.

        Raises:
            ValueError: If the request is invalid.
        """
        if self._job_template is None:
            raise ValueError('"jobTemplate" is required and cannot be None')

        self._job_template.validate()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from setuptools import find_packages, setup


try:
    with open("README.md") as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = "Apache Gravitino Python client"

setup(
    name="apache-gravitino",
    description="Python lib/client for Apache Gravitino",
    version="1.2.0rc4",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Apache Software Foundation",
    author_email="dev@gravitino.apache.org",
    maintainer="Apache Gravitino Community",
    maintainer_email="dev@gravitino.apache.org",
    license="Apache-2.0",
    url="https://github.com/apache/gravitino",
    python_requires=">=3.10",
    keywords="Data, AI, metadata, catalog",
    packages=find_packages(exclude=["tests*", "scripts*"]),
    project_urls={
        "Homepage": "https://gravitino.apache.org/",
        "Source Code": "https://github.com/apache/gravitino",
        "Documentation": "https://gravitino.apache.org/docs/overview",
        "Bug Tracker": "https://github.com/apache/gravitino/issues",
        "Slack Chat": "https://the-asf.slack.com/archives/C078RESTT19",
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    install_requires=open("requirements.txt").read(),
    extras_require={
        "dev": open("requirements-dev.txt").read(),
    },
    include_package_data=True,
)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from enum import Enum
from typing import Dict, Optional

from gravitino.api.auditable import Auditable
from gravitino.api.supports_schemas import SupportsSchemas


class Catalog(Auditable):
    """The interface of a catalog. The catalog is the second level entity in the gravitino system,
    containing a set of tables.
    """

    class Type(Enum):
        """The type of the catalog."""

        RELATIONAL = ("relational", False)
        """"Catalog Type for Relational Data Structure, like db.table, catalog.db.table."""

        FILESET = ("fileset", True)
        """Catalog Type for Fileset System (including HDFS, S3, etc.), like path/to/file"""

        MESSAGING = ("messaging", False)
        """Catalog Type for Message Queue, like kafka://topic"""

        MODEL = ("model", True)
        """Catalog Type for ML model"""

        UNSUPPORTED = ("unsupported", False)
        """Catalog Type for test only."""

        def __init__(self, type_name, supports_managed_catalog):
            self._type_name = type_name
            self._supports_managed_catalog = supports_managed_catalog

        @classmethod
        def type_serialize(cls, catalog_type):
            return catalog_type.type_name

        @classmethod
        def type_deserialize(cls, type_name):
            for member in cls:
                if member.type_name == type_name:
                    return member
            return cls.UNSUPPORTED

        @property
        def supports_managed_catalog(self):
            """
            A flag to indicate if the catalog type supports managed catalog. Managed catalog is a
            concept in Gravitino, which means Gravitino will manage the lifecycle of the catalog
            and its subsidiaries. If the catalog type supports managed catalog, users can create
            managed catalog of this type without specifying the catalog provider, Gravitino will
            use the type as the provider to create the managed catalog. If the catalog type does
            not support managed catalog, users need to specify the provider to create the catalog.
            """
            return self._supports_managed_catalog

        @property
        def type_name(self):
            """
            The name of the catalog type.
            """
            return self._type_name

    PROPERTY_PACKAGE = "package"
    """A reserved property to specify the package location of the catalog. The "package" is a string
    of path to the folder where all the catalog related dependencies is located. The dependencies
    under the "package" will be loaded by Gravitino to create the catalog.
    
    The property "package" is not needed if the catalog is a built-in one, Gravitino will search
    the proper location using "provider" to load the dependencies. Only when the folder is in
    different location, the "package" property is needed.
    """

    @abstractmethod
    def name(self) -> str:
        """
        Returns:
            The name of the catalog.
        """
        pass

    @abstractmethod
    def type(self) -> Type:
        """
        Returns:
            The type of the catalog.
        """
        pass

    @abstractmethod
    def provider(self) -> str:
        """
        Returns:
            The provider of the catalog.
        """
        pass

    @abstractmethod
    def comment(self) -> Optional[str]:
        """The comment of the catalog. Note. this method will return null if the comment is not set for
        this catalog.

        Returns:
            The provider of the catalog.
        """
        pass

    @abstractmethod
    def properties(self) -> Optional[Dict[str, str]]:
        """
        The properties of the catalog. Note, this method will return null if the properties are not set.

        Returns:
            The properties of the catalog.
        """
        pass

    def as_schemas(self) -> SupportsSchemas:
        """Return the {@link SupportsSchemas} if the catalog supports schema operations.

        Raises:
            UnsupportedOperationException if the catalog does not support schema operations.

        Returns:
            The {@link SupportsSchemas} if the catalog supports schema operations.
        """
        raise UnsupportedOperationException(
            "Catalog does not support schema operations"
        )

    def as_table_catalog(self) -> "TableCatalog":  # noqa: F821
        """
        Raises:
            UnsupportedOperationException if the catalog does not support table operations.

        Returns:
            the {@link TableCatalog} if the catalog supports table operations.
        """
        raise UnsupportedOperationException("Catalog does not support table operations")

    def as_fileset_catalog(self) -> "FilesetCatalog":  # noqa: F821
        """
        Raises:
            UnsupportedOperationException if the catalog does not support fileset operations.

        Returns:
            the FilesetCatalog if the catalog supports fileset operations.
        """
        raise UnsupportedOperationException(
            "Catalog does not support fileset operations"
        )

    def as_topic_catalog(self) -> "TopicCatalog":  # noqa: F821
        """
        Returns:
            the {@link TopicCatalog} if the catalog supports topic operations.

        Raises:
            UnsupportedOperationException if the catalog does not support topic operations.
        """
        raise UnsupportedOperationException("Catalog does not support topic operations")

    def as_model_catalog(self) -> "ModelCatalog":  # noqa: F821
        """
        Returns:
            the {@link ModelCatalog} if the catalog supports model operations.

        Raises:
            UnsupportedOperationException if the catalog does not support model operations.
        """
        raise UnsupportedOperationException("Catalog does not support model operations")

    def as_function_catalog(self) -> "FunctionCatalog":  # noqa: F821
        """
        Returns:
            the {@link FunctionCatalog} if the catalog supports function operations.

        Raises:
            UnsupportedOperationException if the catalog does not support function operations.
        """
        raise UnsupportedOperationException(
            "Catalog does not support function operations"
        )


class UnsupportedOperationException(Exception):
    pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from functools import reduce
from typing import ClassVar, List, Optional

from gravitino.api.rel.indexes.index import Index
from gravitino.utils.precondition import Precondition


class IndexDTO(Index):
    """Data transfer object representing index information.

    Attributes:
        EMPTY_INDEXES (List[IndexDTO]): An empty array of indexes.
    """

    EMPTY_INDEXES: ClassVar[List["IndexDTO"]] = []

    def __init__(
        self,
        index_type: Index.IndexType,
        name: Optional[str],
        field_names: List[List[str]],
    ):
        Precondition.check_argument(index_type is not None, "Index type cannot be null")
        Precondition.check_argument(
            field_names is not None and len(field_names) > 0,
            "The index must be set with corresponding column names",
        )

        self._index_type = index_type
        self._name = name
        self._field_names = field_names

    def type(self) -> Index.IndexType:
        return self._index_type

    def name(self) -> Optional[str]:
        return self._name

    def field_names(self) -> List[List[str]]:
        return self._field_names

    def __eq__(self, other) -> bool:
        if not isinstance(other, IndexDTO):
            return False
        return (
            self._index_type is other.type()
            and self._name == other.name()
            and self._field_names == other.field_names()
        )

    def __hash__(self) -> int:
        initial_hash = hash((self._index_type, self._name))
        return reduce(
            lambda result, field_name: 31 * result + hash(tuple(field_name)),
            self._field_names,
            initial_hash,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from abc import ABC, abstractmethod
from enum import Enum, unique
from typing import List, Optional


class Index(ABC):
    """The Index interface defines methods for implementing table index columns.

    Currently, settings for `PRIMARY_KEY` and `UNIQUE_KEY` are provided.
    """

    @unique
    class IndexType(str, Enum):
        """The enum IndexType defines the type of the index.

        Currently, `PRIMARY_KEY` and `UNIQUE_KEY` are supported.
        """

        PRIMARY_KEY = "PRIMARY_KEY"
        """`PRIMARY` KEY index in a relational database is a field or a combination of fields that
        uniquely identifies each record in a table. It serves as a unique identifier for each row,
        ensuring that no two rows have the same key. The PRIMARY KEY is used to establish
        relationships between tables and enforce the entity integrity of a database. Additionally, it
        helps in indexing and organizing the data for efficient retrieval and maintenance."""

        UNIQUE_KEY = "UNIQUE_KEY"
        """UNIQUE KEY in a relational database is a field or a combination of fields that ensures each
        record in a table has a distinct value or combination of values. Unlike a primary key, a
        UNIQUE KEY allows for the presence of null values, but it still enforces the constraint that
        no two records can have the same unique key value(s). UNIQUE KEYs are used to maintain data
        integrity by preventing duplicate entries in specific columns, and they can be applied to
        columns that are not designated as the primary key. The uniqueness constraint imposed by
        UNIQUE KEY helps in avoiding redundancy and ensuring data accuracy in the database."""

    @abstractmethod
    def type(self) -> IndexType:
        """Returns the type of the index. eg: `PRIMARY_KEY` and `UNIQUE_KEY`.

        Returns:
            IndexType: The type of the index. eg: `PRIMARY_KEY` and `UNIQUE_KEY`.
        """
        pass  # pragma: no cover

    @abstractmethod
    def name(self) -> Optional[str]:
        """Returns the name of the index.

        Returns:
            Optional[str]: The name of the index.
        """
        pass  # pragma: no cover

    @abstractmethod
    def field_names(self) -> List[List[str]]:
        """Returns the field name under the table contained in the index.

        It is the column names, could be "a.b.c" for nested column, but normally it
        could only be "a".

        Returns:
            List[List[str]]: The field name under the table contained in the index.
        """
        pass  # pragma: no cover



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from gravitino.api.rel.expressions.expression import Expression


class NamedReference(Expression):
    """
    Represents a field or column reference in the public logical expression API.
    """

    @staticmethod
    def field(field_name: list[str]) -> FieldReference:
        """
        Returns a FieldReference for the given field name(s). The array of field name(s) is
        used to reference nested fields. For example, if we have a struct column named "student" with a
        data type of StructType{"name": StringType, "age": IntegerType}, we can reference the field
        "name" by calling field("student", "name").

        @param field_name the field name(s)
        @return a FieldReference for the given field name(s)
        """
        return FieldReference(field_name)

    @staticmethod
    def field_from_column(column_name: str) -> FieldReference:
        """Returns a FieldReference for the given column name."""
        return FieldReference([column_name])

    def field_name(self) -> list[str]:
        """
        Returns the referenced field name as a list of string parts.
        Must be implemented by subclasses.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def children(self) -> list[Expression]:
        """Named references do not have children."""
        return Expression.EMPTY_EXPRESSION

    def references(self) -> list[NamedReference]:
        """Named references reference themselves."""
        return [self]


class FieldReference(NamedReference):
    """
    A NamedReference that references a field or column.
    """

    _field_names: list[str]

    def __init__(self, field_names: list[str]):
        super().__init__()
        self._field_names = field_names

    def field_name(self) -> list[str]:
        return self._field_names

    def __eq__(self, other: object) -> bool:
        if isinstance(other, FieldReference):
            return self._field_names == other._field_names
        return False

    def __hash__(self) -> int:
        return hash(tuple(self._field_names))

    def __str__(self) -> str:
        """Returns the string representation of the field reference."""
        return ".".join(self._field_names)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements. See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership. The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

from copy import deepcopy
from typing import List, Dict, Optional, Any
from .job_template import JobTemplate, JobType


class SparkJobTemplate(JobTemplate):
    """
    Represents a job template for executing Spark jobs.

    Note: The class_name field is required for Java/Scala Spark applications but is
    optional for PySpark applications. For PySpark jobs, this field can be None.
    """

    def __init__(self, builder: "SparkJobTemplate.Builder"):
        """
        Initializes a SparkJobTemplate instance.
        """
        super().__init__(builder)
        self._class_name = builder.class_name
        self._jars = builder.jars
        self._files = builder.files
        self._archives = builder.archives
        self._configs = builder.configs

    @property
    def class_name(self) -> str:
        """
        Returns the class name of the Spark job.

        This field is required for Java/Scala Spark applications but optional for PySpark
        applications. For PySpark jobs, this may return None.

        Returns:
            the class name as a string, or None for PySpark applications.
        """
        return self._class_name

    @property
    def jars(self) -> List[str]:
        """
        Returns the list of JAR files associated with this Spark job template.

        Returns:
            List of JAR file paths.
        """
        return self._jars

    @property
    def files(self) -> List[str]:
        """
        Returns the list of files associated with this Spark job template.

        Returns:
            List of file paths.
        """
        return self._files

    @property
    def archives(self) -> List[str]:
        """
        Returns the list of archives associated with this Spark job template.

        Returns:
            List of archive file paths.
        """
        return self._archives

    @property
    def configs(self) -> Dict[str, str]:
        """
        Returns the configuration settings for this Spark job template.

        Returns:
            A dictionary of configuration key-value pairs.
        """
        return self._configs

    def job_type(self) -> JobType:
        """
        Returns the type of the job, which is SPARK for this template.

        Returns:
            JobType.SPARK: Indicates that this is a Spark job template.
        """
        return JobType.SPARK

    @staticmethod
    def builder() -> "SparkJobTemplate.Builder":
        """
        Creates a new builder instance for constructing a SparkJobTemplate.

        Returns:
            SparkJobTemplate.Builder: A new builder instance.
        """
        return SparkJobTemplate.Builder()

    def __eq__(self, other: Any) -> bool:
        if not super().__eq__(other):
            return False
        return (
            self._class_name == other.class_name
            and self._jars == other.jars
            and self._files == other.files
            and self._archives == other.archives
            and self._configs == other.configs
        )

    def __hash__(self) -> int:
        return hash(
            (
                super().__hash__(),
                self._class_name,
                tuple(self._jars),
                tuple(self._files),
                tuple(self._archives),
                frozenset(self._configs.items()),
            )
        )

    def __str__(self) -> str:
        result = "\nSparkJobTemplate{\n"
        result += f"  className='{self.class_name}',\n"

        if self._jars:
            result += (
                "  jars=[\n" + "".join(f"    {j},\n" for j in self._jars) + "  ],\n"
            )
        else:
            result += "  jars=[],\n"

        if self._files:
            result += (
                "  files=[\n" + "".join(f"    {f},\n" for f in self._files) + "  ],\n"
            )
        else:
            result += "  files=[],\n"

        if self._archives:
            result += (
                "  archives=[\n"
                + "".join(f"    {a},\n" for a in self._archives)
                + "  ],\n"
            )
        else:
            result += "  archives=[],\n"

        if self._configs:
            result += (
                "  configs={\n"
                + "".join(f"    {k}: {v},\n" for k, v in self.configs.items())
                + "  },\n"
            )
        else:
            result += "  configs={},\n"

        return result + super().__str__() + "}\n"

    class Builder(
        JobTemplate.BaseBuilder["SparkJobTemplate.Builder", "SparkJobTemplate"]
    ):
        """
        Builder class for creating SparkJobTemplate instances.
        """

        def __init__(self):
            """
            Initializes the builder with default values.
            """
            super().__init__()
            self.class_name: Optional[str] = None
            self.jars: List[str] = []
            self.files: List[str] = []
            self.archives: List[str] = []
            self.configs: Dict[str, str] = {}

        def with_class_name(self, class_name: str) -> "SparkJobTemplate.Builder":
            """
            Sets the class name for this Spark job template.

            This field is required for Java/Scala Spark applications but optional for PySpark
            applications. For PySpark jobs, you can omit this call or set it to None.

            Args:
                class_name: The fully qualified name of the Spark job class, or None for PySpark applications.

            Returns:
                The builder instance for method chaining.
            """
            self.class_name = class_name
            return self

        def with_jars(self, jars: List[str]) -> "SparkJobTemplate.Builder":
            """
            Sets the JAR files for this Spark job template.

            Args:
                jars: A list of JAR file paths to be included in the Spark job.

            Returns:
                The builder instance for method chaining.
            """
            self.jars = deepcopy(jars)
            return self

        def with_files(self, files: List[str]) -> "SparkJobTemplate.Builder":
            """
            Sets the files for this Spark job template.

            Args:
                files: A list of file paths to be included in the Spark job.

            Returns:
                The builder instance for method chaining.
            """
            self.files = deepcopy(files)
            return self

        def with_archives(self, archives: List[str]) -> "SparkJobTemplate.Builder":
            """
            Sets the archives for this Spark job template.

            Args:
                archives: A list of archive file paths to be included in the Spark job.

            Returns:
                The builder instance for method chaining.
            """
            self.archives = deepcopy(archives)
            return self

        def with_configs(self, configs: Dict[str, str]) -> "SparkJobTemplate.Builder":
            """
            Sets the configuration settings for this Spark job template.

            Args:
                configs: A dictionary of configuration key-value pairs to be applied to the Spark job.

            Returns:
                The builder instance for method chaining.
            """
            self.configs = deepcopy(configs)
            return self

        def validate(self):
            """
            Validates the SparkJobTemplate properties.

            Note: className is required for Java/Scala Spark applications but optional for PySpark.
            """
            super().validate()
            # className is optional - required for Java/Scala but not for PySpark
            self.jars = self.jars or []
            self.files = self.files or []
            self.archives = self.archives or []
            self.configs = self.configs or {}

        def build(self) -> "SparkJobTemplate":
            """
            Builds and returns a SparkJobTemplate instance.
            """
            self.validate()
            return SparkJobTemplate(self)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import ABC, abstractmethod

from gravitino.api.tag.tag import Tag


class SupportsTags(ABC):
    """Interface for supporting getting or associate tags to objects.

    This interface will be mixed with metadata objects to provide tag operations.
    """

    @abstractmethod
    def list_tags(self) -> list[str]:
        """List all the tag names for the specific object.

        Returns:
            list[str]: The list of tag names.
        """
        pass

    @abstractmethod
    def list_tags_info(self) -> list[Tag]:
        """List all the tags with details for the specific object.

        Returns:
            list[Tag]: The list of tags.
        """
        pass

    @abstractmethod
    def get_tag(self, name: str) -> Tag:
        """Get a tag by its name for the specific object.

        Args:
            name (str): The name of the tag.

        Raises:
            NoSuchTagException: If the tag does not associate with the object.

        Returns:
            Tag: The tag.
        """
        pass

    @abstractmethod
    def associate_tags(
        self, tags_to_add: list[str], tags_to_remove: list[str]
    ) -> list[str]:
        """Associate tags to the specific object.

        The `tags_to_add` will be added to the object, and the `tags_to_remove` will be removed from the object.

        Note that:
        1. Adding or removing tags that are not existed will be ignored.
        2. If the same name tag is in both `tags_to_add` and `tags_to_remove`, it will be ignored.
        3. If the tag is already associated with the object, it will raise `TagAlreadyAssociatedException`.

        Args:
            tags_to_add (list[str]): The arrays of tag name to be added to the object.
            tags_to_remove (list[str]): The array of tag name to be removed from the object.

        Raises:
            TagAlreadyAssociatedException: If the tag is already associated with the object.

        Returns:
            list[str]: The array of tag names that are associated with the object.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from gravitino.api.job.job_handle import JobHandle
from gravitino.dto.job.job_dto import JobDTO


class GenericJobHandle(JobHandle):
    """Represents a generic job handle."""

    def __init__(self, job_dto: JobDTO):
        self._job_dto = job_dto

    def job_template_name(self) -> str:
        return self._job_dto.job_template_name()

    def job_id(self) -> str:
        return self._job_dto.job_id()

    def job_status(self):
        return self._job_dto.status()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from typing import Dict

from gravitino.api.credential.credential import Credential
from gravitino.utils.precondition import Precondition


class ADLSTokenCredential(Credential, ABC):
    """Represents ADLS token credential."""

    ADLS_TOKEN_CREDENTIAL_TYPE: str = "adls-token"
    ADLS_DOMAIN: str = "dfs.core.windows.net"
    _STORAGE_ACCOUNT_NAME: str = "azure-storage-account-name"
    _SAS_TOKEN: str = "adls-sas-token"

    def __init__(self, credential_info: Dict[str, str], expire_time_in_ms: int):
        self._account_name = credential_info.get(self._STORAGE_ACCOUNT_NAME, None)
        self._sas_token = credential_info.get(self._SAS_TOKEN, None)
        self._expire_time_in_ms = expire_time_in_ms
        Precondition.check_string_not_empty(
            self._account_name, "The ADLS account name should not be empty."
        )
        Precondition.check_string_not_empty(
            self._sas_token, "The ADLS SAS token should not be empty."
        )
        Precondition.check_argument(
            self._expire_time_in_ms > 0,
            "The expiration time of ADLS token credential should be greater than 0",
        )

    def credential_type(self) -> str:
        """The type of the credential.

        Returns:
             the type of the credential.
        """
        return self.ADLS_TOKEN_CREDENTIAL_TYPE

    def expire_time_in_ms(self) -> int:
        """Returns the expiration time of the credential in milliseconds since
        the epoch, 0 means it will never expire.

        Returns:
             The expiration time of the credential.
        """
        return self._expire_time_in_ms

    def credential_info(self) -> Dict[str, str]:
        """The credential information.

        Returns:
             The credential information.
        """
        return {
            self._STORAGE_ACCOUNT_NAME: self._account_name,
            self._SAS_TOKEN: self._sas_token,
        }

    def account_name(self) -> str:
        """The ADLS account name.

        Returns:
            The ADLS account name.
        """
        return self._account_name

    def sas_token(self) -> str:
        """The ADLS sas token.

        Returns:
            The ADLS sas token.
        """
        return self._sas_token



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC


class ModelChange(ABC):
    """
    A model change is a change to a model. It can be used to rename a model, update the comment
    of a model, set a property and value pair for a model, or remove a property from a model.
    """

    @staticmethod
    def rename(new_name):
        """Creates a new model change to rename the model.
        Args:
            new_name: The new name of the model.
        Returns:
            The model change.
        """
        return ModelChange.RenameModel(new_name)

    @staticmethod
    def set_property(pro, value):
        """Creates a new model change to set the property and value pairs for the model.
        Args:
            property: The name of the property to be set.
            value: The value of the property to be set.
        Returns:
            The model change.
        """
        return ModelChange.SetProperty(pro, value)

    @staticmethod
    def remove_property(pro):
        """Creates a new model change to remove the property and value pairs for the model.
        Args:
            property: The name of the property to be removed.
        Returns:
            The model change.
        """
        return ModelChange.RemoveProperty(pro)

    @staticmethod
    def update_comment(comment):
        """Creates a new model change to update the comment of the model.
        Args:
            comment: The new comment of the model.
        Returns:
            The model change.
        """
        return ModelChange.UpdateComment(comment)

    class RenameModel:
        """A model change to rename the model."""

        def __init__(self, new_name):
            self._new_name = new_name

        def new_name(self):
            """Retrieves the new name set for the model.
            Returns:
                The new name of the model.
            """
            return self._new_name

        def __eq__(self, other) -> bool:
            """Compares this RenameModel instance with another object for equality. Two instances are
            considered equal if they designate the same new name for the model.

            Args:
                other: The object to compare with this instance.

            Returns:
                true if the given object represents an identical model renaming operation; false otherwise.
            """
            if not isinstance(other, ModelChange.RenameModel):
                return False
            return self.new_name() == other.new_name()

        def __hash__(self):
            """Generates a hash code for this RenameModel instance. The hash code is primarily based on
            the new name for the model.

            Returns:
                A hash code value for this renaming operation.
            """
            return hash(self.new_name())

        def __str__(self):
            """Provides a string representation of the RenameModel instance. This string includes the
            class name followed by the new name of the model.

            Returns:
                A string summary of this renaming operation.
            """
            return f"RENAMEMODEL {self.new_name()}"

    class SetProperty:
        """
        A model change to set the property and value pairs for the model.
        """

        def __init__(self, pro, value):
            self._property = pro
            self._value = value

        def property(self):
            """Retrieves the name of the property to be set.
            Returns:
                The name of the property.
            """
            return self._property

        def value(self):
            """Retrieves the value of the property to be set.
            Returns:
                The value of the property.
            """
            return self._value

        def __eq__(self, other) -> bool:
            """Compares this SetProperty instance with another object for equality. Two instances are
            considered equal if they designate the same property and value for the model.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model property setting operation; false otherwise.
            """
            if not isinstance(other, ModelChange.SetProperty):
                return False
            return self.property() == other.property() and self.value() == other.value()

        def __hash__(self):
            """Generates a hash code for this SetProperty instance. The hash code is primarily based on
            the property and value for the model.
            Returns:
                A hash code value for this property setting operation.
            """
            return hash(self.property(), self.value())

        def __str__(self):
            """Provides a string representation of the SetProperty instance. This string includes the
            class name followed by the property and value of the model.
            Returns:
                A string summary of this property setting operation.
            """
            return f"SETPROPERTY {self.property()}={self.value()}"

    class RemoveProperty:
        """
        A model change to remove the property and value pairs for the model.
        """

        def __init__(self, pro):
            self._property = pro

        def property(self):
            """Retrieves the name of the property to be removed.
            Returns:
                The name of the property.
            """
            return self._property

        def __eq__(self, other) -> bool:
            """Compares this RemoveProperty instance with another object for equality. Two instances are
            considered equal if they designate the same property for the model.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model property removal operation; false otherwise.
            """
            if not isinstance(other, ModelChange.RemoveProperty):
                return False
            return self.property() == other.property()

        def __hash__(self):
            """Generates a hash code for this RemoveProperty instance. The hash code is primarily based on
            the property for the model.
            Returns:
                A hash code value for this property removal operation.
            """
            return hash(self.property())

        def __str__(self):
            """Provides a string representation of the RemoveProperty instance. This string includes the
            class name followed by the property of the model.
            Returns:
                A string summary of this property removal operation.
            """
            return f"REMOVEPROPERTY {self.property()}"

    class UpdateComment:
        """
        A model change to update the comment of the model.
        """

        def __init__(self, new_comment):
            self._new_comment = new_comment

        def new_comment(self):
            """Retrieves the comment of the model.
            Returns:
                The comment of the model.
            """
            return self._new_comment

        def __eq__(self, other) -> bool:
            """Compares this UpdateComment instance with another object for equality. Two instances are
            considered equal if they designate the same comment for the model.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model comment update operation; false otherwise.
            """
            if not isinstance(other, ModelChange.UpdateComment):
                return False
            return self.new_comment() == other.new_comment()

        def __hash__(self):
            """Generates a hash code for this UpdateComment instance. The hash code is primarily based on
            the comment for the model.
            Returns:
                A hash code value for this comment update operation.
            """
            return hash(self.new_comment())

        def __str__(self):
            """Provides a string representation of the UpdateComment instance. This string includes the
            class name followed by the comment of the model.
            Returns:
                A string summary of this comment update operation.
            """
            return f"UpdateComment {self.new_comment()}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from abc import abstractmethod

from gravitino.api.auditable import Auditable
from gravitino.api.authorization.securable_objects import SecurableObject


class Role(Auditable):
    """The interface of a role. The role is the entity which has kinds of privileges. One role can have
    multiple privileges of multiple securable objects.
    """

    @abstractmethod
    def name(self) -> str:
        """
        The name of the role.

        Raises:
            NotImplementedError: if the method is not implemented.

        Returns:
            str: The name of the role.
        """
        raise NotImplementedError()

    @abstractmethod
    def properties(self) -> dict[str, str]:
        """
        The properties of the role. Note, this method will return None if the properties are not set.

        Raises:
            NotImplementedError: if the method is not implemented.

        Returns:
            dict[str, str]: The properties of the role.
        """
        raise NotImplementedError()

    @abstractmethod
    def securable_objects(self) -> list[SecurableObject]:
        """
        The securable object represents a special kind of entity with a unique identifier. All
        securable objects are organized by tree structure. For example: If the securable object is a
        table, the identifier may be `catalog1.schema1.table1`._summary_

        Raises:
            NotImplementedError: if the method is not implemented.

        Returns:
            list[SecurableObject]: The securable objects of the role.
        """
        raise NotImplementedError()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.audit import Audit


@dataclass
class AuditDTO(Audit, DataClassJsonMixin):
    """Data transfer object representing audit information."""

    _creator: Optional[str] = field(default=None, metadata=config(field_name="creator"))
    """The creator of the audit."""

    _create_time: Optional[str] = field(
        default=None, metadata=config(field_name="createTime")
    )  # TODO: Can't deserialized datetime from JSON
    """The create time of the audit."""

    _last_modifier: Optional[str] = field(
        default=None, metadata=config(field_name="lastModifier")
    )
    """The last modifier of the audit."""

    _last_modified_time: Optional[str] = field(
        default=None, metadata=config(field_name="lastModifiedTime")
    )  # TODO: Can't deserialized datetime from JSON
    """The last modified time of the audit."""

    def __hash__(self):
        return hash(
            (
                self.creator(),
                self.create_time(),
                self.last_modifier(),
                self.last_modified_time(),
            )
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, AuditDTO):
            return False
        return (
            self.creator() == other.creator()
            and self.create_time() == other.create_time()
            and self.last_modifier() == other.last_modifier()
            and self.last_modified_time() == other.last_modified_time()
        )

    def creator(self) -> str:
        """The creator of the entity.

        Returns:
             the creator of the entity.
        """
        return self._creator

    def create_time(self) -> str:
        """The creation time of the entity.

        Returns:
             The creation time of the entity.
        """
        return self._create_time

    def last_modifier(self) -> str:
        """
        Returns:
             The last modifier of the entity.
        """
        return self._last_modifier

    def last_modified_time(self) -> str:
        """
        Returns:
             The last modified time of the entity.
        """
        return self._last_modified_time



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, Optional

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.function.function_impl import FunctionImpl
from gravitino.api.function.java_impl import JavaImpl
from gravitino.api.function.python_impl import PythonImpl
from gravitino.api.function.sql_impl import SQLImpl
from gravitino.dto.function.function_resources_dto import FunctionResourcesDTO


class FunctionImplDTO(DataClassJsonMixin, ABC):
    """Abstract DTO for function implementation."""

    @abstractmethod
    def language(self) -> FunctionImpl.Language:
        """Returns the language of this implementation."""
        pass

    @abstractmethod
    def to_function_impl(self) -> FunctionImpl:
        """Convert this DTO to a FunctionImpl instance."""
        pass


@dataclass
class SQLImplDTO(FunctionImplDTO):
    """DTO for SQL function implementation."""

    _runtime: str = field(metadata=config(field_name="runtime"))
    _sql: str = field(metadata=config(field_name="sql"))
    _language: str = field(default="SQL", metadata=config(field_name="language"))
    _resources: Optional[FunctionResourcesDTO] = field(
        default=None, metadata=config(field_name="resources")
    )
    _properties: Optional[Dict[str, str]] = field(
        default=None, metadata=config(field_name="properties")
    )

    def language(self) -> FunctionImpl.Language:
        return FunctionImpl.Language.SQL

    def runtime(self) -> str:
        return self._runtime

    def sql(self) -> str:
        return self._sql

    def resources(self) -> Optional[FunctionResourcesDTO]:
        return self._resources

    def properties(self) -> Optional[Dict[str, str]]:
        return self._properties

    def to_function_impl(self) -> FunctionImpl:
        return SQLImpl(
            runtime=FunctionImpl.RuntimeType.from_string(self._runtime),
            sql=self._sql,
            resources=(
                self._resources.to_function_resources() if self._resources else None
            ),
            properties=self._properties,
        )

    @classmethod
    def from_function_impl(cls, impl) -> "SQLImplDTO":
        """Create a SQLImplDTO from a SQLImpl instance."""
        return cls(
            _runtime=impl.runtime().name,
            _sql=impl.sql(),
            _resources=FunctionResourcesDTO.from_function_resources(impl.resources()),
            _properties=impl.properties() if impl.properties() else None,
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, SQLImplDTO):
            return False
        return (
            self._runtime == other._runtime
            and self._sql == other._sql
            and self._resources == other._resources
            and self._properties == other._properties
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._runtime,
                self._sql,
                self._resources,
                frozenset(self._properties.items()) if self._properties else None,
            )
        )


@dataclass
class JavaImplDTO(FunctionImplDTO):
    """DTO for Java function implementation."""

    _runtime: str = field(metadata=config(field_name="runtime"))
    _class_name: str = field(metadata=config(field_name="className"))
    _language: str = field(default="JAVA", metadata=config(field_name="language"))
    _resources: Optional[FunctionResourcesDTO] = field(
        default=None, metadata=config(field_name="resources")
    )
    _properties: Optional[Dict[str, str]] = field(
        default=None, metadata=config(field_name="properties")
    )

    def language(self) -> FunctionImpl.Language:
        return FunctionImpl.Language.JAVA

    def runtime(self) -> str:
        return self._runtime

    def class_name(self) -> str:
        return self._class_name

    def resources(self) -> Optional[FunctionResourcesDTO]:
        return self._resources

    def properties(self) -> Optional[Dict[str, str]]:
        return self._properties

    def to_function_impl(self) -> FunctionImpl:
        return JavaImpl(
            runtime=FunctionImpl.RuntimeType.from_string(self._runtime),
            class_name=self._class_name,
            resources=(
                self._resources.to_function_resources() if self._resources else None
            ),
            properties=self._properties,
        )

    @classmethod
    def from_function_impl(cls, impl) -> "JavaImplDTO":
        """Create a JavaImplDTO from a JavaImpl instance."""
        return cls(
            _runtime=impl.runtime().name,
            _class_name=impl.class_name(),
            _resources=FunctionResourcesDTO.from_function_resources(impl.resources()),
            _properties=impl.properties() if impl.properties() else None,
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, JavaImplDTO):
            return False
        return (
            self._runtime == other._runtime
            and self._class_name == other._class_name
            and self._resources == other._resources
            and self._properties == other._properties
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._runtime,
                self._class_name,
                self._resources,
                frozenset(self._properties.items()) if self._properties else None,
            )
        )


@dataclass
class PythonImplDTO(FunctionImplDTO):
    """DTO for Python function implementation."""

    _runtime: str = field(metadata=config(field_name="runtime"))
    _handler: str = field(metadata=config(field_name="handler"))
    _language: str = field(default="PYTHON", metadata=config(field_name="language"))
    _code_block: Optional[str] = field(
        default=None, metadata=config(field_name="codeBlock")
    )
    _resources: Optional[FunctionResourcesDTO] = field(
        default=None, metadata=config(field_name="resources")
    )
    _properties: Optional[Dict[str, str]] = field(
        default=None, metadata=config(field_name="properties")
    )

    def language(self) -> FunctionImpl.Language:
        return FunctionImpl.Language.PYTHON

    def runtime(self) -> str:
        return self._runtime

    def handler(self) -> str:
        return self._handler

    def code_block(self) -> Optional[str]:
        return self._code_block

    def resources(self) -> Optional[FunctionResourcesDTO]:
        return self._resources

    def properties(self) -> Optional[Dict[str, str]]:
        return self._properties

    def to_function_impl(self) -> FunctionImpl:
        return PythonImpl(
            runtime=FunctionImpl.RuntimeType.from_string(self._runtime),
            handler=self._handler,
            code_block=self._code_block,
            resources=(
                self._resources.to_function_resources() if self._resources else None
            ),
            properties=self._properties,
        )

    @classmethod
    def from_function_impl(cls, impl) -> "PythonImplDTO":
        """Create a PythonImplDTO from a PythonImpl instance."""
        return cls(
            _runtime=impl.runtime().name,
            _handler=impl.handler(),
            _code_block=impl.code_block(),
            _resources=FunctionResourcesDTO.from_function_resources(impl.resources()),
            _properties=impl.properties() if impl.properties() else None,
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, PythonImplDTO):
            return False
        return (
            self._runtime == other._runtime
            and self._handler == other._handler
            and self._code_block == other._code_block
            and self._resources == other._resources
            and self._properties == other._properties
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._runtime,
                self._handler,
                self._code_block,
                self._resources,
                frozenset(self._properties.items()) if self._properties else None,
            )
        )


def function_impl_dto_from_function_impl(impl: FunctionImpl) -> FunctionImplDTO:
    """Create a FunctionImplDTO from a FunctionImpl instance."""
    if isinstance(impl, SQLImpl):
        return SQLImplDTO.from_function_impl(impl)
    if isinstance(impl, JavaImpl):
        return JavaImplDTO.from_function_impl(impl)
    if isinstance(impl, PythonImpl):
        return PythonImplDTO.from_function_impl(impl)

    raise ValueError(f"Unsupported implementation type: {type(impl)}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, Optional

from gravitino.api.function.function import Function
from gravitino.api.function.function_definition import FunctionDefinition
from gravitino.api.function.function_type import FunctionType
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.function.function_dto import FunctionDTO


class GenericFunction(Function):
    """A generic implementation of the Function interface."""

    def __init__(self, function_dto: FunctionDTO):
        """Create a GenericFunction from a FunctionDTO.

        Args:
            function_dto: The function DTO.
        """
        self._function_dto = function_dto

    def name(self) -> str:
        """Returns the function name."""
        return self._function_dto.name()

    def function_type(self) -> FunctionType:
        """Returns the function type."""
        return self._function_dto.function_type()

    def deterministic(self) -> bool:
        """Returns whether the function is deterministic."""
        return self._function_dto.deterministic()

    def comment(self) -> Optional[str]:
        """Returns the optional comment of the function."""
        return self._function_dto.comment()

    def definitions(self) -> List[FunctionDefinition]:
        """Returns the definitions of the function."""
        return self._function_dto.definitions()

    def audit_info(self) -> Optional[AuditDTO]:
        """Returns the audit information."""
        return self._function_dto.audit_info()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from gravitino.constants.error import ErrorConstants
from gravitino.dto.responses.error_response import ErrorResponse
from gravitino.exceptions.handlers.rest_error_handler import RestErrorHandler
from gravitino.exceptions.base import (
    CatalogNotInUseException,
    NoSuchCredentialException,
)


class CredentialErrorHandler(RestErrorHandler):
    def handle(self, error_response: ErrorResponse):
        error_message = error_response.format_error_message()
        code = error_response.code()
        exception_type = error_response.type()

        if code == ErrorConstants.NOT_FOUND_CODE:
            if exception_type == NoSuchCredentialException.__name__:
                raise NoSuchCredentialException(error_message)

        if code == ErrorConstants.NOT_IN_USE_CODE:
            raise CatalogNotInUseException(error_message)

        super().handle(error_response)


CREDENTIAL_ERROR_HANDLER = CredentialErrorHandler()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import TypeVar

from gravitino.constants.timeout import TIMEOUT

T = TypeVar("T")


class GravitinoClientConfig:
    """
    Configuration class for Gravitino Python client;
    It encapsulates HTTP connection configurations and validates client properties.
    """

    POSITIVE_NUMBER_ERROR_MSG: str = "The value must be a positive number"
    """Error message for positive number validation"""

    GRAVITINO_CLIENT_CONFIG_PREFIX: str = "gravitino_client_"
    """Configuration key prefix for Gravitino client config"""

    CLIENT_REQUEST_TIMEOUT_DEFAULT: int = TIMEOUT
    """Default HTTP request timeout in seconds"""

    CLIENT_REQUEST_TIMEOUT: str = "gravitino_client_request_timeout"
    """Configuration key for request timeout"""

    SUPPORT_CLIENT_CONFIG_KEYS: set = {CLIENT_REQUEST_TIMEOUT}
    """Set of supported configuration keys"""

    def __init__(self, properties: dict):
        """Initializes the configuration with validated properties

        Args:
            properties: Key-value pairs of configuration parameters
        """
        self._properties = properties

    @classmethod
    def build_from_properties(cls, properties: dict) -> "GravitinoClientConfig":
        """Factory method to create configuration from properties

        Args:
            properties: Input configuration dictionary

        Returns:
            GravitinoClientConfig instance

        Raises:
            ValueError: If unsupported keys are provided
        """
        if properties is None:
            return cls({})
        for key in properties:
            if key not in cls.SUPPORT_CLIENT_CONFIG_KEYS:
                raise ValueError(f"Invalid property for client: {key}")
        return cls({key: properties[key] for key in properties})

    def get_client_request_timeout(self) -> int:
        """Retrieves and validates HTTP request timeout

        Returns:
            Timeout value in seconds

        Raises:
            IllegalArgumentException: If value is negative or value cannot be converted to integer
        """
        timeout = self._property_as_int(
            self.CLIENT_REQUEST_TIMEOUT, self.CLIENT_REQUEST_TIMEOUT_DEFAULT
        )
        self._check_value(
            self.CLIENT_REQUEST_TIMEOUT,
            timeout,
            lambda x: x >= 0,
            self.POSITIVE_NUMBER_ERROR_MSG,
        )
        return timeout

    def _property_as_int(self, key: str, default: int) -> int:
        """Safely converts property value to integer

        Args:
            key: Configuration key to retrieve
            default: Default value if key not found

        Returns:
            Converted integer value

        Raises:
            ValueError: If value cannot be converted to integer
        """
        value = self._properties.get(key)
        if value is None:
            return default
        try:
            return int(value)
        except (TypeError, ValueError) as err:
            raise ValueError(
                f"Value '{value}' for key '{key}' must be an integer"
            ) from err

    def _check_value(self, key: str, value: T, validator: callable, error_msg: str):
        """Generic validation method for configuration values

        Args:
            key: Config key
            value: Value to validate
            validator: Validation function returning boolean
            error_msg: Error message template

        Raises:
            ValueError: If validation fails
        """
        if not validator(value):
            raise ValueError(f"Value '{value}' for key '{key}' is invalid. {error_msg}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.api.job.job_template import JobTemplate, JobType
from gravitino.api.job.job_template_change import (
    JobTemplateChange,
    RenameJobTemplate,
    ShellTemplateUpdate,
    SparkTemplateUpdate,
    TemplateUpdate,
    UpdateJobTemplate,
    UpdateJobTemplateComment,
)
from gravitino.api.job.shell_job_template import ShellJobTemplate
from gravitino.api.job.spark_job_template import SparkJobTemplate
from gravitino.api.metalake_change import MetalakeChange
from gravitino.client.fileset_catalog import FilesetCatalog
from gravitino.client.generic_model_catalog import GenericModelCatalog
from gravitino.client.relational_catalog import RelationalCatalog
from gravitino.dto.catalog_dto import CatalogDTO
from gravitino.dto.job.job_template_dto import JobTemplateDTO
from gravitino.dto.job.shell_job_template_dto import ShellJobTemplateDTO
from gravitino.dto.job.shell_template_update_dto import ShellTemplateUpdateDTO
from gravitino.dto.job.spark_job_template_dto import SparkJobTemplateDTO
from gravitino.dto.job.spark_template_update_dto import SparkTemplateUpdateDTO
from gravitino.dto.job.template_update_dto import TemplateUpdateDTO
from gravitino.dto.requests.catalog_update_request import CatalogUpdateRequest
from gravitino.dto.requests.job_template_update_request import (
    JobTemplateUpdateRequest,
    RenameJobTemplateRequest,
    UpdateJobTemplateCommentRequest,
    UpdateJobTemplateContentRequest,
)
from gravitino.dto.requests.metalake_update_request import MetalakeUpdateRequest
from gravitino.namespace import Namespace
from gravitino.utils import HTTPClient


class DTOConverters:
    """Utility class for converting between DTOs and domain objects."""

    @staticmethod
    def to_metalake_update_request(change: MetalakeChange) -> object:
        # Assuming MetalakeUpdateRequest has similar nested class structure for requests
        if isinstance(change, MetalakeChange.RenameMetalake):
            return MetalakeUpdateRequest.RenameMetalakeRequest(change.new_name())
        if isinstance(change, MetalakeChange.UpdateMetalakeComment):
            return MetalakeUpdateRequest.UpdateMetalakeCommentRequest(
                change.new_comment()
            )
        if isinstance(change, MetalakeChange.SetProperty):
            return MetalakeUpdateRequest.SetMetalakePropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, MetalakeChange.RemoveProperty):
            return MetalakeUpdateRequest.RemoveMetalakePropertyRequest(
                change.property()
            )

        raise ValueError(f"Unknown change type: {type(change).__name__}")

    @staticmethod
    def to_catalog(metalake: str, catalog: CatalogDTO, client: HTTPClient):
        namespace = Namespace.of(metalake)
        if catalog.type() == Catalog.Type.FILESET:
            return FilesetCatalog(
                namespace=namespace,
                name=catalog.name(),
                catalog_type=catalog.type(),
                provider=catalog.provider(),
                comment=catalog.comment(),
                properties=catalog.properties(),
                audit=catalog.audit_info(),
                rest_client=client,
            )

        if catalog.type() == Catalog.Type.MODEL:
            return GenericModelCatalog(
                namespace=namespace,
                name=catalog.name(),
                catalog_type=catalog.type(),
                provider=catalog.provider(),
                comment=catalog.comment(),
                properties=catalog.properties(),
                audit=catalog.audit_info(),
                rest_client=client,
            )
        if catalog.type() == Catalog.Type.RELATIONAL:
            return RelationalCatalog(
                catalog_namespace=namespace,
                name=catalog.name(),
                catalog_type=catalog.type(),
                provider=catalog.provider(),
                comment=catalog.comment(),
                properties=catalog.properties(),
                audit=catalog.audit_info(),
                rest_client=client,
            )

        raise NotImplementedError("Unsupported catalog type: " + str(catalog.type()))

    @staticmethod
    def to_catalog_update_request(change: CatalogChange):
        if isinstance(change, CatalogChange.RenameCatalog):
            return CatalogUpdateRequest.RenameCatalogRequest(change.new_name())
        if isinstance(change, CatalogChange.UpdateCatalogComment):
            return CatalogUpdateRequest.UpdateCatalogCommentRequest(
                change.new_comment()
            )
        if isinstance(change, CatalogChange.SetProperty):
            # TODO
            # pylint: disable=too-many-function-args
            return CatalogUpdateRequest.SetCatalogPropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, CatalogChange.RemoveProperty):
            return CatalogUpdateRequest.RemoveCatalogPropertyRequest(
                change.get_property()
            )

        raise ValueError(f"Unknown change type: {type(change).__name__}")

    @staticmethod
    def _to_shell_job_template_dto(dto: JobTemplateDTO) -> ShellJobTemplateDTO:
        if isinstance(dto, ShellJobTemplateDTO):
            return dto

        raise TypeError(
            f"The provided JobTemplateDTO is not of type ShellJobTemplateDTO: {type(dto)}"
        )

    @staticmethod
    def _to_spark_job_template_dto(dto: JobTemplateDTO) -> SparkJobTemplateDTO:
        if isinstance(dto, SparkJobTemplateDTO):
            return dto

        raise TypeError(
            f"The provided JobTemplateDTO is not of type SparkJobTemplateDTO: {type(dto)}"
        )

    @staticmethod
    def to_shell_job_template(template: JobTemplate) -> ShellJobTemplate:
        if isinstance(template, ShellJobTemplate):
            return template

        raise TypeError(
            f"The provided JobTemplate is not of type ShellJobTemplate: {type(template)}"
        )

    @staticmethod
    def to_spark_job_template(template: JobTemplate) -> SparkJobTemplate:
        if isinstance(template, SparkJobTemplate):
            return template

        raise TypeError(
            f"The provided JobTemplate is not of type SparkJobTemplate: {type(template)}"
        )

    @staticmethod
    def from_job_template_dto(dto: JobTemplateDTO) -> JobTemplate:
        if dto.job_type() == JobType.SHELL:
            shell_dto = DTOConverters._to_shell_job_template_dto(dto)
            return (
                ShellJobTemplate.builder()
                .with_name(shell_dto.name())
                .with_comment(shell_dto.comment())
                .with_executable(shell_dto.executable())
                .with_arguments(shell_dto.arguments())
                .with_environments(shell_dto.environments())
                .with_custom_fields(shell_dto.custom_fields())
                .with_scripts(shell_dto.scripts())
                .build()
            )

        if dto.job_type() == JobType.SPARK:
            spark_dto = DTOConverters._to_spark_job_template_dto(dto)
            return (
                SparkJobTemplate.builder()
                .with_name(spark_dto.name())
                .with_comment(spark_dto.comment())
                .with_executable(spark_dto.executable())
                .with_arguments(spark_dto.arguments())
                .with_environments(spark_dto.environments())
                .with_custom_fields(spark_dto.custom_fields())
                .with_class_name(spark_dto.class_name())
                .with_jars(spark_dto.jars())
                .with_files(spark_dto.files())
                .with_archives(spark_dto.archives())
                .with_configs(spark_dto.configs())
                .build()
            )

        raise ValueError(f"Unsupported job type: {dto.job_type()}")

    @staticmethod
    def to_job_template_dto(template: JobTemplate) -> JobTemplateDTO:
        if isinstance(template, ShellJobTemplate):
            shell_template = DTOConverters.to_shell_job_template(template)
            return ShellJobTemplateDTO(
                _job_type=shell_template.job_type(),
                _name=shell_template.name,
                _comment=shell_template.comment,
                _executable=shell_template.executable,
                _arguments=shell_template.arguments,
                _environments=shell_template.environments,
                _custom_fields=shell_template.custom_fields,
                _scripts=shell_template.scripts,
            )

        if isinstance(template, SparkJobTemplate):
            spark_template = DTOConverters.to_spark_job_template(template)
            return SparkJobTemplateDTO(
                _job_type=spark_template.job_type(),
                _name=spark_template.name,
                _comment=spark_template.comment,
                _executable=spark_template.executable,
                _arguments=spark_template.arguments,
                _environments=spark_template.environments,
                _custom_fields=spark_template.custom_fields,
                _class_name=spark_template.class_name,
                _jars=spark_template.jars,
                _files=spark_template.files,
                _archives=spark_template.archives,
                _configs=spark_template.configs,
            )

        raise ValueError(f"Unsupported job type: {type(template)}")

    @staticmethod
    def to_template_update_dto(template_update: TemplateUpdate) -> TemplateUpdateDTO:
        if isinstance(template_update, ShellTemplateUpdate):
            return ShellTemplateUpdateDTO(
                new_executable=template_update.get_new_executable(),
                new_arguments=template_update.get_new_arguments(),
                new_environments=template_update.get_new_environments(),
                new_custom_fields=template_update.get_new_custom_fields(),
                new_scripts=template_update.get_new_scripts(),
            )

        if isinstance(template_update, SparkTemplateUpdate):
            return SparkTemplateUpdateDTO(
                new_executable=template_update.get_new_executable(),
                new_arguments=template_update.get_new_arguments(),
                new_environments=template_update.get_new_environments(),
                new_custom_fields=template_update.get_new_custom_fields(),
                new_class_name=template_update.get_new_class_name(),
                new_jars=template_update.get_new_jars(),
                new_files=template_update.get_new_files(),
                new_archives=template_update.get_new_archives(),
                new_configs=template_update.get_new_configs(),
            )

        raise ValueError(f"Unsupported template update type: {type(template_update)}")

    @staticmethod
    def to_job_template_update_request(
        change: JobTemplateChange,
    ) -> JobTemplateUpdateRequest:
        if isinstance(change, RenameJobTemplate):
            return RenameJobTemplateRequest(change.get_new_name())

        if isinstance(change, UpdateJobTemplateComment):
            return UpdateJobTemplateCommentRequest(change.get_new_comment())

        if isinstance(change, UpdateJobTemplate):
            template_update_dto = DTOConverters.to_template_update_dto(
                change.get_template_update()
            )
            return UpdateJobTemplateContentRequest(template_update_dto)

        raise ValueError(f"Unknown change type: {type(change).__name__}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from gravitino.api.rel.expressions.unparsed_expression import UnparsedExpression
from gravitino.dto.rel.expressions.function_arg import FunctionArg


class UnparsedExpressionDTO(UnparsedExpression, FunctionArg):
    """Data transfer object representing an unparsed expression."""

    def __init__(self, unparsed_expression: str):
        self._unparsed_expression = unparsed_expression

    def unparsed_expression(self) -> str:
        """Returns the unparsed expression as a string.

        Returns:
            str: The value of the unparsed expression.
        """
        return self._unparsed_expression

    def arg_type(self) -> FunctionArg.ArgType:
        return self.ArgType.UNPARSED

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UnparsedExpressionDTO):
            return False
        return self._unparsed_expression == other.unparsed_expression()

    def __hash__(self) -> int:
        return hash((self.arg_type(), self._unparsed_expression))

    def __str__(self) -> str:
        return (
            f"UnparsedExpressionDTO{{unparsedExpression='{self._unparsed_expression}'}}"
        )

    @staticmethod
    def builder() -> Builder:
        """A builder instance for `UnparsedExpressionDTO`.

        Returns:
            Builder: A builder instance for `UnparsedExpressionDTO`.
        """
        return UnparsedExpressionDTO.Builder()

    class Builder:
        """Builder for `UnparsedExpressionDTO`."""

        def __init__(self):
            self._unparsed_expression = None

        def with_unparsed_expression(
            self, unparsed_expression: str
        ) -> UnparsedExpressionDTO.Builder:
            """Set the unparsed expression.

            Args:
                unparsed_expression (str): The unparsed expression.

            Returns:
                Builder: The builder.
            """

            self._unparsed_expression = unparsed_expression
            return self

        def build(self) -> UnparsedExpressionDTO:
            """Build the unparsed expression.

            Returns:
                UnparsedExpressionDTO: The unparsed expression.
            """

            return UnparsedExpressionDTO(unparsed_expression=self._unparsed_expression)
